import pandas as pd
from sklearn.base import BaseEstimator
from sklearn.base import TransformerMixin


class CUSUMFilter(BaseEstimator, TransformerMixin):
    """
    Downsamples a timeseries to filter out non significant value
    changing samples.

    :param threshold: Sets the threshold to trigger a sample, setting
                      the threshold higher will result in less samples being selected.
    :type threshold: float
    """

    def __init__(self, threshold: float):
        self._threshold = threshold

    def transform(self, X: pd.Series) -> pd.DatetimeIndex:
        """
        Returns a pandas DatetimeIndex indicating which samples have been
        selected by the CUSUM filter.

        :param X: Time series with time indices.
        :type X: pd.Series
        :return: DatetimeIndex indicating which samples have been selected.
        :rtype: pd.DatetimeIndex
        """
        filtered_indices = []
        pos_threshold = 0
        neg_threshold = 0
        diff = X.diff()
        for index in diff.index[1:]:
            pos_threshold, neg_threshold = (
                max(0, pos_threshold + diff.loc[index]),
                min(0, neg_threshold + diff.loc[index]),
            )

            if neg_threshold < -self._threshold:
                neg_threshold = 0
                filtered_indices.append(index)

            elif pos_threshold > self._threshold:
                pos_threshold = 0
                filtered_indices.append(index)

        return pd.DatetimeIndex(filtered_indices)



from abc import abstractmethod
from typing import Tuple

import numpy as np
import pandas as pd
from numba import jit
from numba import prange
from sklearn.base import BaseEstimator
from sklearn.base import TransformerMixin

from mizarlabs.static import CLOSE
from mizarlabs.static import DAILY_VOL
from mizarlabs.static import EVENT_END_TIME
from mizarlabs.static import LABEL
from mizarlabs.static import PROFIT_TAKING
from mizarlabs.static import RETURN
from mizarlabs.static import SIDE
from mizarlabs.static import STOP_LOSS
from mizarlabs.static import TIMESTAMP_UNIT
from mizarlabs.transformers.utils import check_missing_columns
from mizarlabs.transformers.utils import convert_to_timestamp

__all__ = [
    "BaseLabeling",
    "TripleBarrierMethodLabeling",
    "triple_barrier_labeling",
    "get_labels",
    "get_daily_vol",
]


class BaseLabeling(BaseEstimator, TransformerMixin):
    """Base class for labeling."""

    def __init__(self, n_expiration_bars: int):
        self.n_expiration_bars = n_expiration_bars

    def fit(self, y):
        """
        Fit the model (just for sklearn compatibility).

        :param x:
        :param y:
        :return:
        """
        return self

    def transform(self, y: pd.DataFrame) -> pd.DataFrame:
        """
        Return a dataframe with the target labelled.

        :param y:
        :return:
        """
        self._check_input(y)
        output = self._transform(y)
        self._check_output(output)
        return output

    @abstractmethod
    def _check_output(self, output_df: pd.DataFrame):
        pass

    @abstractmethod
    def _check_input(self, input_series: pd.DataFrame):
        pass

    @abstractmethod
    def _transform(self, y: pd.DataFrame) -> pd.DataFrame:
        pass


# TODO: think about a triple barrier labeling with no
#  volatility adjustment
class TripleBarrierMethodLabeling(BaseLabeling):
    """
    Implements the triple barrier method used to label the target.

    See page 45 of Advances in Financial Machine Learning by Marcos Lopez de Prado
    for additional information.

    :param num_expiration_bars: Max number of bars from the position taking to the position closing.
    :type num_expiration_bars: int
    :param profit_taking_factor: The factor that multiplies the volatility for
        the creation of the horizontal upper barrier
    :type profit_taking_factor: float
    :param stop_loss_factor: The factor that multiplies the volatility for the
        creation of the horizontal lower barrier
    :type stop_loss_factor: float
    :param metalabeling: Whether metalabeling is activated
    :type metalabeling: bool, optional
    :param close_column_name: The name of the close column
    :type close_column_name: str, optional
    :param side_column_name: The name of the side column (metalabeling)
    :type side_column_name: str, optional
    :param volatility_window: The number of bars used for the volatility calculation
    :type volatility_window: int, optional
    :param expiration_label: Labels with 0 are returned to indicate expiration / vertical barrier has been hit
    :type expiration_label: bool, optional
    """

    def __init__(
        self,
        num_expiration_bars: int,
        profit_taking_factor: float,
        stop_loss_factor: float,
        metalabeling: bool = False,
        close_column_name: str = CLOSE,
        side_column_name: str = SIDE,
        volatility_window: int = 100,
        expiration_label: bool = False,
    ):
        super().__init__(num_expiration_bars)
        self.profit_taking_factor = profit_taking_factor
        self.stop_loss_factor = stop_loss_factor
        self.metalabeling = metalabeling
        self.close_column_name = close_column_name
        self.side_column_name = side_column_name
        self.volatility_window = volatility_window
        self.expiration_label = expiration_label
        if self.metalabeling:
            assert (
                self.side_column_name
            ), "Need to to set side_column_name when meta-labeling is selected"

    def fit(self, X, y=None, **fit_params):
        return self

    def _check_output(self, output_df: pd.DataFrame):
        check_missing_columns(output_df, [EVENT_END_TIME, DAILY_VOL, RETURN, LABEL])

    def _check_input(self, input_df: pd.DataFrame):
        if self.metalabeling:
            check_missing_columns(
                input_df, [self.close_column_name, self.side_column_name]
            )
        else:
            check_missing_columns(input_df, [self.close_column_name])

        assert isinstance(input_df, pd.DataFrame), (
            "Please provide only a Pandas dataframe with close prices and"
            " side (in case of metalabeling) with datetime indices."
        )

    def _transform(self, y: pd.DataFrame) -> pd.DataFrame:
        """
        Return a dataframe with the target labelled.

        :param y: The dataframe with the close and side
        :type y: pd.DataFrame
        :return: Dataframe containing labeled target
        :rtype: pd.DataFrame
        """
        barriers_info_df = pd.DataFrame(index=y.index)

        barriers_info_df[EVENT_END_TIME] = (
            pd.Series(barriers_info_df.index).shift(-self.n_expiration_bars).values
        )

        barriers_info_df[DAILY_VOL] = get_daily_vol(
            y[self.close_column_name], self.volatility_window
        )
        if SIDE not in y.columns:
            barriers_info_df[self.side_column_name] = 1
        else:
            barriers_info_df[self.side_column_name] = y[self.side_column_name]

        barriers_df = triple_barrier_labeling(
            close=y[self.close_column_name],
            barrier_info_df=barriers_info_df,
            profit_taking_factor=self.profit_taking_factor,
            stop_loss_factor=self.stop_loss_factor,
        )

        barriers_info_df[EVENT_END_TIME] = barriers_df.dropna(how="all").min(axis=1)

        bins = get_labels(
            barriers_df,
            barriers_info_df,
            y[self.close_column_name],
            metalabeling=self.metalabeling,
            expiration_label=self.expiration_label,
        )

        barriers_info_df.drop(SIDE, axis=1, inplace=True)

        output_df = pd.merge(
            barriers_info_df[[EVENT_END_TIME, DAILY_VOL]],
            bins,
            left_index=True,
            right_index=True,
            how="left",
        )

        output_df.loc[output_df[DAILY_VOL].isna(), EVENT_END_TIME] = pd.NaT
        output_df.loc[output_df[DAILY_VOL].isna(), RETURN] = np.nan
        output_df.loc[output_df[DAILY_VOL].isna(), LABEL] = np.nan
        return output_df


def triple_barrier_labeling(
    close: pd.Series,
    barrier_info_df: pd.DataFrame,
    profit_taking_factor: float,
    stop_loss_factor: float,
) -> pd.DataFrame:
    """
    Calculate the first hit on the stop loss and profit taking barrier.

    As described in Advances in financial machine learning,
    Marcos Lopez de Prado, 2018.

    :param close: Series of prices.
    :type close: pd.Series
    :param barrier_info_df: Info for creating the barriers
    :type barrier_info_df: pd.DataFrame
    :param profit_taking_factor: The factor that multiplies the volatility for
        the creation of the horizontal upper barrier
    :type profit_taking_factor: float
    :param stop_loss_factor: The factor that multiplies the volatility for the
        creation of the horizontal lower barrier
    :type stop_loss_factor: float
    :return: Dataframe containing the first hit for each of the barriers
    :rtype: pd.DataFrame
    """
    check_missing_columns(barrier_info_df, [EVENT_END_TIME, DAILY_VOL, SIDE])

    if profit_taking_factor < 0 and stop_loss_factor < 0:
        raise ValueError("Stop loss and profit taking factors should be greater than 0")

    barriers_df = barrier_info_df[[EVENT_END_TIME]].copy(deep=True)

    # Creating the profit taking barriers. If the profit taking factor is 0
    # then there won't be any profit taking barriers.
    if profit_taking_factor > 0:
        profit_taking_barriers = profit_taking_factor * barrier_info_df[DAILY_VOL]
    else:
        profit_taking_barriers = pd.Series(index=barrier_info_df.index)  # NaNs

    # Creating the stop loss barriers. If the stop loss factor is 0
    # then there won't be any stop loss barriers.
    if stop_loss_factor > 0:
        stop_loss_barriers = -stop_loss_factor * barrier_info_df[DAILY_VOL]
    else:
        stop_loss_barriers = pd.Series(index=barrier_info_df.index)  # NaNs

    (
        profit_taking_barriers_timestamps,
        stop_loss_barriers_timestamps,
    ) = get_horizontal_barriers_hit(
        convert_to_timestamp(
            barrier_info_df[EVENT_END_TIME].fillna(close.index[-1]).values
        ),
        convert_to_timestamp(barrier_info_df.index.values),
        close.values,
        barrier_info_df[SIDE].values,
        stop_loss_barriers.values,
        profit_taking_barriers.values,
    )

    profit_taking_barriers_timestamps = pd.Series(
        profit_taking_barriers_timestamps,
        dtype=pd.Int64Dtype(),
        index=barriers_df.index,
    ).replace(0, np.nan)

    stop_loss_barriers_timestamps = pd.Series(
        stop_loss_barriers_timestamps, dtype=pd.Int64Dtype(), index=barriers_df.index
    ).replace(0, np.nan)

    barriers_df.loc[:, STOP_LOSS] = pd.to_datetime(
        stop_loss_barriers_timestamps, unit=TIMESTAMP_UNIT
    )
    barriers_df.loc[:, PROFIT_TAKING] = pd.to_datetime(
        profit_taking_barriers_timestamps, unit=TIMESTAMP_UNIT
    )

    assert set(barriers_df.columns) == {EVENT_END_TIME, STOP_LOSS, PROFIT_TAKING}

    return barriers_df


@jit(parallel=True, nopython=True)
def get_horizontal_barriers_hit(
    expiration_barrier_timestamps: np.ndarray,
    index_timestamps: np.ndarray,
    close: np.ndarray,
    side: np.ndarray,
    stop_loss_barriers: np.ndarray,
    profit_taking_barriers: np.ndarray,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Get the horizontal barriers hits.

    :param expiration_barrier_timestamps: Array with the timestamps of the vertical barrier
    :type expiration_barrier_timestamps: np.ndarray
    :param index_timestamps: Array with the timestamps of the position taking
    :type index_timestamps: np.ndarray
    :param close: Array with the price of the close
    :type close: np.ndarray
    :param side: Array with the side of the position (1, -1)
    :type side: np.ndarray
    :param stop_loss_barriers: Array with the stop loss barriers values
    :type stop_loss_barriers: np.ndarray
    :param profit_taking_barriers: Array with teh profit taking barriers values
    :type profit_taking_barriers: np.ndarray
    :return: Two arrays containing when the profit taking and the stop loss
             barriers have been hit
    :rtype: Tuple[np.ndarray, np.ndarray]
    """

    stop_loss_barriers_timestamps = np.zeros_like(
        expiration_barrier_timestamps, dtype=np.float64
    )
    profit_taking_barriers_timestamps = np.zeros_like(
        expiration_barrier_timestamps, dtype=np.float64
    )

    for i in prange(len(index_timestamps)):
        # creating mask between the position taking date and expiration barrier
        # date
        within_barrier_mask = np.greater_equal(
            index_timestamps, index_timestamps[i]
        ) * np.less(index_timestamps, expiration_barrier_timestamps[i])

        # selecting all the prices between the position taking date
        # and expiration barrier date
        close_prices_within_barrier = close[within_barrier_mask]

        # Calculating returns within the prices in the barriers
        returns_within_barrier = (close_prices_within_barrier / close[i] - 1) * side[i]

        # Calculating the first date when the stop loss barrier is hit.
        # If the barrier is not hit, the result is nan
        stop_loss_mask = np.less(returns_within_barrier, stop_loss_barriers[i])

        if np.any(stop_loss_mask):
            stop_loss_barriers_timestamps[i] = np.min(
                index_timestamps[within_barrier_mask][stop_loss_mask]
            )

        # Calculating the first date when the profit taking barrier is hit.
        # If the barrier is not hit, the result is nan
        profit_taking_mask = np.greater(
            returns_within_barrier, profit_taking_barriers[i]
        )
        if np.any(profit_taking_mask):
            profit_taking_barriers_timestamps[i] = np.min(
                index_timestamps[within_barrier_mask][profit_taking_mask]
            )

    return profit_taking_barriers_timestamps, stop_loss_barriers_timestamps


def get_labels(
    barriers_df: pd.DataFrame,
    barriers_info_df: pd.DataFrame,
    close: pd.Series,
    metalabeling: bool,
    expiration_label: bool = False,
) -> pd.DataFrame:
    """
    Calculate returns and assign return classes based on the first touched bar.

    Case 1: ('side' not in barriers_info_df): bin in (-1,1) <-label by price action
    Case 2: ('side' in barriers_info_df): bin in (0,1) <-label by pnl (meta-labeling)

    :param barriers_df: dataframe with datetime when barriers are hit
    :type barriers_df: pd.DataFrame
    :param barriers_info_df: Info for creating the barriers
    :type barriers_info_df: pd.DataFrame
    :param close: Series of prices.
    :type close: pd.Series
    :param metalabeling: Whether or not metalabelign is activated
    :type metalabeling: bool

    :return: Dataframe containing event
    :rtype: pd.DataFrame
    """
    check_missing_columns(barriers_info_df, [EVENT_END_TIME, DAILY_VOL, SIDE])

    # selecting bars that have a closed position
    events_without_na = barriers_info_df.dropna(subset=[EVENT_END_TIME])
    list_existing_dates = events_without_na.index.union(
        events_without_na[EVENT_END_TIME].values
    ).drop_duplicates()
    px = close.reindex(list_existing_dates, method="bfill")

    # creating dataframe with start position date
    output_df = pd.DataFrame(index=events_without_na.index)
    # calculate the returns
    output_df[RETURN] = (
        px.loc[events_without_na[EVENT_END_TIME].values].values
        / px.loc[events_without_na.index]
        - 1
    )
    returns = (
        close.loc[events_without_na[EVENT_END_TIME]].values
        / close.loc[events_without_na.index]
        - 1
    ).rename(RETURN)

    pd.testing.assert_series_equal(returns, output_df[RETURN])

    if metalabeling:
        # metalabeling labels are always only 1 and 0. 1 is assigned when the
        # base model is correct while 0 is assigned when the base model is not
        # correct. When the base model does not take a position or the returns
        #  are 0 then the metalabel will be na
        output_df[RETURN] *= events_without_na[SIDE]
        output_df.loc[output_df[RETURN] > 0, LABEL] = 1
        output_df.loc[output_df[RETURN] < 0, LABEL] = 0
        output_df.loc[output_df[RETURN] == 0, LABEL] = np.nan
        # When metalabeling is not activated then the labels can be 1 or -1.
        # 1 is when the returns are positive and -1 when the returns are
        # negative
    else:
        output_df[LABEL] = np.sign(output_df[RETURN])

    if expiration_label:
        expired_events = barriers_df.loc[events_without_na.index].loc[
            barriers_df.stop_loss.isna() & barriers_df.profit_taking.isna()
        ]
        output_df.loc[expired_events.index, LABEL] = 0

    assert set(output_df.columns) == {RETURN, LABEL}

    return output_df


def get_daily_vol(close: pd.Series, ewm_span: int = 100) -> pd.Series:
    """
    Estimate the daily volatility.

    :param close: Contains the close price
    :type close: pd.Series
    :param ewm_span: The span of the standard deviation
    :type ewm_span: int
    :return: The daily volatility per each bar
    :rtype: pd.Series
    """
    # find nearest 1 day apart close price
    indices_with_date_diff = close.index.searchsorted(
        close.index - pd.Timedelta(days=1)
    )
    indices_filtered = indices_with_date_diff[indices_with_date_diff > 0]

    # creating series with the date of the nearest 1 day apart  close price
    shifted_close = pd.Series(
        close.index[indices_filtered - 1],
        index=close.index[close.shape[0] - indices_filtered.shape[0] :],
    )
    # calculating returns between close and nearest 1 day apart  close
    returns = close.loc[shifted_close.index] / close.loc[shifted_close].values - 1
    daily_vol = returns.ewm(span=ewm_span).std()
    return daily_vol



import numpy as np
import pandas as pd
from scipy import sparse
from typing import Union, Tuple


def get_ind_matrix(
    samples_info_sets: pd.Series,
    price_bars: pd.DataFrame,
    event_end_time_column_name: str,
    return_indices: bool = False,
) -> Union[sparse.lil_matrix, Tuple[sparse.lil_matrix, pd.Timestamp]]:
    """
    Snippet 4.3, page 65, Build an Indicator Matrix
    Get indicator matrix. The book implementation uses bar_index as input,
    however there is no explanation how to form it. We decided that
    using triple_barrier_events and price bars by analogy with concurrency
    is the best option.

    :param samples_info_sets: Series indicating the start and end time of a sample, e.g. from triple barrier
    :type samples_info_sets: pd.Series
    :param price_bars: Price bars which were used to form triple barrier events or other labelling method
    :type price_bars: pd.DataFrame
    :param event_end_time_column_name: Column name
    :type event_end_time_column_name: str
    :return: Indicator binary matrix indicating what (price) bars influence the label for each observation and in
             addition in can also return the respective timestamp indices
    :rtype: Union[sparse.lil_matrix, Tuple[sparse.lil_matrix, pd.Timestamp]]
    """
    if (
        bool(samples_info_sets.isnull().values.any()) is True
        or bool(samples_info_sets.index.isnull().any()) is True
    ):
        raise ValueError("NaN values in triple_barrier_events, delete nans")

    triple_barrier_events = pd.DataFrame(
        samples_info_sets
    )  # Convert Series to DataFrame
    # Take only period covered in triple_barrier_events
    trimmed_price_bars_index = price_bars[
        (price_bars.index >= triple_barrier_events.index.min())
        & (price_bars.index <= triple_barrier_events[event_end_time_column_name].max())
    ].index

    label_endtime = triple_barrier_events[event_end_time_column_name]
    bar_index = sorted(
        list(
            {
                *triple_barrier_events.index,
                *triple_barrier_events[event_end_time_column_name],
                *trimmed_price_bars_index,
            }
        )
    )  # Drop duplicates and sort

    # Get sorted timestamps with index in sorted array
    sorted_timestamps = dict(zip(sorted(bar_index), range(len(bar_index))))

    tokenized_endtimes = np.column_stack(
        (
            label_endtime.index.map(sorted_timestamps),
            label_endtime.map(sorted_timestamps).values,
        )
    )  # Create array of arrays: [label_index_position, label_endtime_position]
    ind_mat = sparse.lil_matrix(
        (len(bar_index), len(label_endtime))
    )  # Init indicator matrix
    for sample_num, (label_index, label_endtime_index) in enumerate(tokenized_endtimes):
        ind_mat[label_index : label_endtime_index + 1, sample_num] = 1

    ind_mat_csc = ind_mat.tocsc()
    assert np.max(ind_mat_csc) == 1
    assert np.min(ind_mat_csc) == 0

    if return_indices:
        return ind_mat, bar_index
    else:
        return ind_mat


def calc_average_uniqueness(ind_mat_csc: sparse.csc_matrix) -> np.ndarray:
    """
    Calculates the average uniqueness of an indicator matrix.

    :param ind_mat_csc: indicator matrix, with size (T x N), where T is no. of timestamps and N the number of samples.
    :type ind_mat_csc: sparse.csc_matrix
    :return: array with average uniqueness per column in the indicator matrix, where a column represents a sample
    :rtype: np.ndarray
    """
    concurrent_events = ind_mat_csc.sum(axis=1)
    uniqueness_matrix = ind_mat_csc.multiply(1 / concurrent_events)
    counts = np.diff(uniqueness_matrix.tocsc().indptr)
    sums = uniqueness_matrix.tocsr().sum(axis=0).A1
    average_uniqueness_array = sums / counts
    return average_uniqueness_array


def _calc_update_avg_unique(
    ind_mat: sparse.csc_matrix,
    samples_to_update: np.ndarray,
    bootstrapped_samples: np.ndarray,
) -> np.ndarray:
    """
    Calculates the average uniqueness for a subset of samples if they were to be added to the currently bootstrap
    samples.

    :param ind_mat: indicator matrix, with size (T x N), where T is no. of timestamps and N the number of samples.
    :type ind_mat: sparse.csc_matrix
    :param samples_to_update: an arrray with indices for the indicator matrix to slice from
    :type samples_to_update: np.ndarray
    :param bootstrapped_samples: an arrray with indices representing the all the bootstrapped samples
    :type bootstrapped_samples: np.ndarray
    :return: array with average uniqueness per sample to update
    :rtype: np.ndarray
    """
    ind_mat_samples = ind_mat[:, samples_to_update]
    partial_conc_events = ind_mat[:, bootstrapped_samples].sum(axis=1)
    non_zero_partial_conc_events_idx = partial_conc_events.nonzero()[0]
    conc_events_per_sample = ind_mat_samples.todense()
    conc_events_per_sample[non_zero_partial_conc_events_idx, :] = (
        conc_events_per_sample[non_zero_partial_conc_events_idx, :]
        + partial_conc_events[non_zero_partial_conc_events_idx]
    )

    # NOTE: this type conversion is needed to do division, but is slow
    conc_events_per_sample = conc_events_per_sample.astype(np.float32)
    conc_events_per_sample[non_zero_partial_conc_events_idx, :] = (
        1 / conc_events_per_sample[non_zero_partial_conc_events_idx]
    )

    uniqueness_matrix_per_sample = ind_mat_samples.multiply(conc_events_per_sample)
    uniqueness_matrix_per_sample_csr = uniqueness_matrix_per_sample.tocsr()
    uniqueness_sums_per_sample = uniqueness_matrix_per_sample_csr.sum(axis=0).A1
    counts_per_sample = np.diff(uniqueness_matrix_per_sample_csr.tocsc().indptr)
    average_uniqueness_array_per_sample = uniqueness_sums_per_sample / counts_per_sample
    return average_uniqueness_array_per_sample


def seq_bootstrap(
    ind_mat: sparse.csc_matrix,
    sample_length: int = None,
    random_state: np.random.RandomState = None,
    update_probs_every: int = 1,
) -> np.array:
    """
    Returns a numpy array with tokenized indices of selected samples,
    which have been selected by sequential bootstrap procedure.

    :param ind_mat: indicator matrix from triple barrier events
    :type ind_mat: sparse.csc_matrix
    :param sample_length: Length of bootstrapped sample, defaults to None
    :type sample_length: int, optional
    :param random_state: random state, defaults to np.random.RandomState()
    :type random_state: np.random.RandomState, optional
    :return: numpy array with tokenized indices of selected samples
    :rtype: np.array
    """
    if random_state is None:
        random_state = np.random.RandomState()

    if sample_length is None:
        sample_length = ind_mat.shape[1]

    bootstrapped_samples = np.array([], dtype=np.intp)  # Bootstrapped samples

    avg_unique = np.ones(ind_mat.shape[1])
    prob = avg_unique / np.sum(avg_unique)
    sample_indices = np.arange(ind_mat.shape[1])
    ind_mat = ind_mat.astype(np.uint32)
    rows_impacted = np.array([], dtype=int)

    for i in range(int(np.ceil(sample_length / update_probs_every))):
        if rows_impacted.shape[0] > 0:
            rows_impacted = np.unique(rows_impacted)
            samples_impacted = ind_mat[rows_impacted, :].nonzero()[1]
            samples_to_update = np.unique(samples_impacted)
            avg_unique[samples_to_update] = _calc_update_avg_unique(
                ind_mat,
                samples_to_update,
                bootstrapped_samples,
            )
            prob = avg_unique / np.sum(avg_unique)  # Draw prob
            rows_impacted = np.array([], dtype=int)

        choices = random_state.choice(
            sample_indices,
            p=prob,
            size=min(sample_length - i * update_probs_every, update_probs_every),
        )
        bootstrapped_samples = np.concatenate((bootstrapped_samples, choices))
        rows_impacted = np.concatenate(
            (rows_impacted, ind_mat[:, choices].nonzero()[0])
        )

    return bootstrapped_samples



"""
Explosiveness tests: SADF
"""
import warnings
from typing import Tuple
from typing import Union

import numpy as np
import pandas as pd
from numba import jit
from numba import prange

from .sdfc import get_beta_and_beta_var


class SupremumAugmentedDickeyFullerStatTest:
    """
    Implementation of SADF, p. 252-261.

    SADF fits the ADF regression at each end point t with backwards expanding start points. For the estimation
    of SADF(t), the right side of the window is fixed at t. SADF recursively expands the beginning of the sample
    up to t - min_length, and returns the sup of this set.

    When doing with sub- or super-martingale test, the variance of beta of a weak long-run bubble may be smaller than
    one of a strong short-run bubble, hence biasing the method towards long-run bubbles. To correct for this bias,
    ADF statistic in samples with large lengths can be penalized with the coefficient phi in [0, 1] such that:

    ADF_penalized = ADF / (t - t0) ^ phi), where t0 is between 1 and t - min_length

    :param model: model test specification, which can be 'no_trend', 'linear',
                  'quadratic', 'sm_poly_1', 'sm_poly_2', 'sm_exp', 'sm_power'
    :type model: str
    :param lags: either number of lags to use or array of specified lags
    :type lags: Union[int, list]
    :param min_length: minimum number of observations needed for estimation
    :type min_length: int
    :param add_constant: flag to add constant, NOTE: sub- and super-martingale tests always
                         have a constant.
    :type add_constant: bool
    :param phi: coefficient to penalize large sample lengths when computing SMT, in [0, 1]
    :type phi: float
    """

    def __init__(
        self,
        model: str,
        lags: Union[int, list],
        min_length: int,
        add_constant: bool,
        phi: float,
    ):
        assert 0 <= phi <= 1, f"Expected phi to be between 0 and 1, got {phi}."
        assert (
            min_length > lags
        ), f"min_length {min_length} should be higher than lags {lags}."
        self.model = model
        self.lags = lags
        self.min_length = min_length
        self.add_constant = add_constant
        self.phi = phi

    def run(self, series_to_test: pd.Series) -> pd.Series:
        """
        Runs the Supremum Augmented Dickey Fuller test and returns a
        series with SADF statistics.

        When the model specification is for sub- and super-martingale
        then a series with SMT statistics is returned. See page 260
        for additional details.

        :param series_to_test: series for which SADF statistics are generated
        :type series_to_test: pd.Series
        :return: series of SADF statistics (or SMT in case of sub- and super-martingale tests)
        :rtype: pd.Series
        """
        X, y = _get_X_y(series_to_test, self.model, self.lags, self.add_constant)
        indices = y.index[self.min_length :]

        sadf_array = _sadf_outer_loop(
            X=X.values.astype(np.float64),
            y=y.values.astype(np.float64).reshape(-1, 1),
            min_length=self.min_length,
            model=self.model,
            phi=self.phi,
        )

        return pd.Series(sadf_array, index=indices)


@jit(parallel=False, nopython=True)
def _get_sadf_at_t(
    X: np.ndarray, y: np.ndarray, min_length: int, model: str, phi: float
) -> np.float64:
    """
    Snippet 17.2, page 258. SADF's Inner Loop (get SADF value at t)

    :param X: matrix of lagged values, constants, trend coefficients
    :type X: np.ndarray
    :param y: y values (either y or y.diff() or np.log(y))
    :type y: np.ndarray
    :param min_length: minimum number of samples needed for estimation
    :type min_length: int
    :param model: either 'no_trend', 'linear', 'quadratic', 'sm_poly_1', 'sm_poly_2', 'sm_exp', 'sm_power'
    :type model: str
    :param phi: coefficient to penalize large sample lengths when computing SMT, in [0, 1]
    :type phi: float
    :return: SADF statistic at time t
    :rtype: np.float64
    """

    adf_array = np.empty(y.shape[0] - min_length + 1)

    # inner loop starts from 1 to t - tau see page 253
    for start in prange(0, y.shape[0] - min_length + 1):
        X_ = np.ascontiguousarray(X[start:])
        y_ = np.ascontiguousarray(y[start:])

        b_mean_, b_var_ = get_beta_and_beta_var(X_, y_)

        current_adf = b_mean_[0, 0] / b_var_[0, 0] ** 0.5

        # if the test specification is a sub- or super-martingale test
        # adjust the test statistic as described on page 260.
        if model[:2] == "sm":
            t = y.shape[0]
            t0 = start + 1  # t0 index starts from 1 to t - tau (page 260)
            current_adf = np.abs(current_adf) / ((y.shape[0] - (t - t0)) ** phi)

        adf_array[start] = current_adf

    return np.max(adf_array)


def _get_X_y_standard_specification(
    series: pd.Series, lags: Union[int, list]
) -> Tuple[pd.DataFrame, pd.Series]:
    """
    Returns the matrix with features X and y for the
    standard specifcation without constant as described on
    page 252.

    :param series: series to calculated SADF values for.
    :type series: pd.Series
    :param lags: either number of lags to use or array of specified lags
    :type lags: Union[int, list]
    :return: matrix with features X and vector y with target values
    :rtype: Tuple[pd.DataFrame, pd.Series]
    """
    series_diff = series.diff().dropna()
    X = _lag_df(series_diff.to_frame(), lags).dropna()
    X["y_lagged"] = series.shift(1).loc[X.index]  # add y_(t-1) column
    y = series_diff.loc[X.index]
    return X, y


def _get_X_y(
    series: pd.Series, model: str, lags: Union[int, list], add_const: bool
) -> Tuple[pd.DataFrame, pd.Series]:
    """
    Snippet 17.2, page 258-259. Preparing The Datasets

    :param series: to prepare for test statistics generation (for example log prices)
    :type series: pd.Series
    :param model: either 'no_trend', 'linear', 'quadratic', 'sm_poly_1', 'sm_poly_2', 'sm_exp', 'sm_power'
    :type model: str
    :param lags: either number of lags to use or array of specified lags
    :type lags: Union[int, list]
    :param add_const: flag to add constant
    :type add_const: bool
    :raises ValueError: if unknown model specification is given.
    :return: prepared X and y for SADF calculation
    :rtype: Tuple[pd.DataFrame, pd.Series]
    """
    if not add_const and model in ["sm_poly_1", "sm_poly_2", "sm_exp", "sm_power"]:
        warnings.warn(
            f"Model specification {model} always has a constant.", category=UserWarning
        )

    if model == "no_trend":
        X, y = _get_X_y_standard_specification(series, lags)
        beta_column = "y_lagged"
        if add_const:
            X["const"] = 1
    elif model == "linear":
        X, y = _get_X_y_standard_specification(series, lags)
        X["trend"] = np.arange(
            1, X.shape[0] + 1
        )  # Add t to the model (0, 1, 2, 3, 4, 5, .... t)
        beta_column = (
            "y_lagged"  # Column which is used to estimate test beta statistics
        )
        if add_const:
            X["const"] = 1
    elif model == "quadratic":
        X, y = _get_X_y_standard_specification(series, lags)
        X["trend"] = np.arange(
            1, X.shape[0] + 1
        )  # Add t to the model (0, 1, 2, 3, 4, 5, .... t)
        X["quad_trend"] = X["trend"] ** 2  # Add t^2 to the model (0, 1, 4, 9, ....)
        beta_column = (
            "y_lagged"  # Column which is used to estimate test beta statistics
        )
        if add_const:
            X["const"] = 1
    elif model == "sm_poly_1":
        y = series.copy()
        X = pd.DataFrame(index=y.index)
        X["const"] = 1
        X["trend"] = np.arange(1, X.shape[0] + 1)
        X["quad_trend"] = X["trend"] ** 2
        beta_column = "quad_trend"
    elif model == "sm_poly_2":
        y = np.log(series.copy())
        X = pd.DataFrame(index=y.index)
        X["const"] = 1
        X["trend"] = np.arange(1, X.shape[0] + 1)
        X["quad_trend"] = X["trend"] ** 2
        beta_column = "quad_trend"
    elif model == "sm_exp":
        y = np.log(series.copy())
        X = pd.DataFrame(index=y.index)
        X["const"] = 1
        X["trend"] = np.arange(1, X.shape[0] + 1)
        beta_column = "trend"
    elif model == "sm_power":
        y = np.log(series.copy())
        X = pd.DataFrame(index=y.index)
        X["const"] = 1
        X["log_trend"] = np.log(np.arange(1, X.shape[0] + 1))
        beta_column = "log_trend"
    else:
        raise ValueError("Unknown model")

    # Move y_lagged column to the front for further extraction
    columns = list(X.columns)
    columns.insert(0, columns.pop(columns.index(beta_column)))
    X = X[columns]

    assert (
        ~X.isna().any().any() and ~y.isna().any().any()
    ), f"The constructed X and y contain NaNs based on the model specification {model}."
    return X, y


def _lag_df(df: pd.DataFrame, lags: Union[int, list]) -> pd.DataFrame:
    """
    Snipet 17.3, page 259. Apply Lags to DataFrame

    :param df: dataframe with features
    :type df: pd.DataFrame
    :param lags: either number of lags to use or array of specified lags
    :type lags: Union[int, list]
    :return: dataframe with features and lagged differenced y features
    :rtype: pd.DataFrame
    """
    df_lagged = pd.DataFrame()
    if isinstance(lags, int):
        lags = range(1, lags + 1)
    else:
        lags = [int(lag) for lag in lags]

    for lag in lags:
        temp_df = df.shift(lag).copy(deep=True)
        temp_df.columns = [str(i) + "_diff_lag_" + str(lag) for i in temp_df.columns]
        df_lagged = df_lagged.join(temp_df, how="outer")
    return df_lagged


@jit(parallel=True, nopython=True)
def _sadf_outer_loop(
    X: np.ndarray,
    y: np.ndarray,
    min_length: np.int64,
    model: str,
    phi: np.float64,
) -> np.ndarray:
    """Runs the SADF outer loop, i.e for each t in T.

    :param X: array with features
    :type X: np.ndarray
    :param y: array with target values
    :type y: np.ndarray
    :param min_length: minimum number of observations
    :type min_length: np.int64
    :param model: either 'no_trend', 'linear', 'quadratic', 'sm_poly_1', 'sm_poly_2', 'sm_exp', 'sm_power'
    :type model: str
    :param phi: coefficient to penalize large sample lengths when computing SMT, in [0, 1]
    :type phi: np.float64
    :return: array with SADF statistics for each point in time t for all T
    :rtype: np.ndarray
    """
    sadf_array = np.empty(y.shape[0] - min_length, dtype=np.float64)
    for index in prange(min_length, y.shape[0]):
        sadf_array[index - min_length] = _get_sadf_at_t(
            X[: (index + 1), :], y[: (index + 1), :], min_length, model, phi
        )
    return sadf_array



from .first_generation import BeckersParkinsonVolatility
from .first_generation import CorwinSchultzSpread
from .first_generation import ParkinsonVolatility
from .first_generation import RollImpact
from .first_generation import RollMeasure
from .second_generation import AmihudLambda
from .second_generation import HasbrouckLambda
from .second_generation import KyleLambda
from .vpin import VPIN



import numpy as np
import pandas as pd
from mizarlabs import static
from mizarlabs.transformers.utils import check_missing_columns
from sklearn.base import BaseEstimator
from sklearn.base import TransformerMixin


class RollMeasure(BaseEstimator, TransformerMixin):
    """
    Implement the roll measure which gives the estimate of effective bid-ask spread.

    See page 282 of Advances in Financial Machine Learning by Marcos Lopez de
    Prado for additional information.

    :param window: The window size
    :type window: int
    :param close_column_name: The name of the close column
    :type close_column_name: str, optional
    """

    def __init__(self, window: int = 20, close_column_name: str = static.CLOSE):
        self.window = window
        self.close_column_name = close_column_name

    def fit(self, X, y=None, **fit_params):
        return self

    def transform(self, X: pd.DataFrame) -> np.ndarray:
        check_missing_columns(X, [self.close_column_name])

        close_diff = X[self.close_column_name].diff()
        close_diff_lag = close_diff.shift(1)
        return (
            2 * np.sqrt(abs(close_diff.rolling(window=self.window).cov(close_diff_lag)))
        ).values.reshape(-1, 1)


class RollImpact(RollMeasure):
    """
    Derivate from Roll Measure which takes into account dollar volume traded.

    See page 282 of Advances in Financial Machine Learning by Marcos Lopez de
    Prado for additional information.

    :param window: The window size
    :type window: int
    :param close_column_name: The name of the close column
    :type close_column_name: str, optional
    :param quote_asset_volume_column_name: The name quote asset column column
    :type quote_asset_volume_column_name: str, optional
    """

    def __init__(
        self,
        window: int = 20,
        close_column_name: str = static.CLOSE,
        quote_asset_volume_column_name: str = static.QUOTE_ASSET_VOLUME,
    ):
        super().__init__(window, close_column_name)
        self.quote_asset_volume_column_name = quote_asset_volume_column_name

    def fit(self, X, y=None, **fit_params):
        return self

    def transform(self, X: pd.DataFrame) -> np.ndarray:
        roll_measure = super().transform(X)
        return roll_measure / X[self.quote_asset_volume_column_name].values.reshape(
            -1, 1
        )


class ParkinsonVolatility(BaseEstimator, TransformerMixin):
    """
    High low volatility estimator developed by Parkinson (1980).

    :param window: The window size
    :type window: int
    :param high_column_name: The name of the high column
    :type high_column_name: str, optional
    :param low_column_name: The name of the low column
    :type low_column_name: str, optional
    """

    def __init__(
        self,
        window: int = 20,
        high_column_name: str = static.HIGH,
        low_column_name: str = static.LOW,
    ):
        self.window = window
        self.high_column_name = high_column_name
        self.low_column_name = low_column_name

    def fit(self, X, y=None, **fit_params):
        return self

    def transform(self, X: pd.DataFrame) -> np.ndarray:
        check_missing_columns(X, [self.high_column_name, self.low_column_name])

        parkinson_constant = 1 / (4 * np.log(2))

        return (
            (
                parkinson_constant
                * np.square(np.log(X[self.high_column_name] / X[self.low_column_name]))
            )
            .rolling(self.window)
            .mean()
        ).values.reshape(-1, 1)


def _get_beta(high: pd.Series, low: pd.Series, window: int) -> pd.Series:
    """
    Get beta estimate from Corwin-Schultz algorithm (p.285, Snippet 19.1).

    :param high: Series containing high prices
    :rtype high: pd.Series
    :param low: Series containing low prices
    :type low: pd.Series
    :param window: The window size of the estimator
    :rtype window: int
    :return: Beta estimates
    :rtype: pd.Series
    """
    ret = np.log(high / low)
    high_low_ret = ret ** 2
    beta = high_low_ret.rolling(window=2).sum()
    beta = beta.rolling(window=window).mean()
    return beta


def _get_gamma(high: pd.Series, low: pd.Series) -> pd.Series:
    """
    Get gamma estimate from Corwin-Schultz algorithm (p.285, Snippet 19.1).

    :param high: Series containing high prices
    :rtype high: pd.Series
    :param low: Series containing low prices
    :type low: pd.Series
    :return: Gamma estimates
    :rtype: pd.Series
    """
    high_max = high.rolling(window=2).max()
    low_min = low.rolling(window=2).min()
    gamma = np.log(high_max / low_min) ** 2
    return gamma


def _get_alpha(beta: pd.Series, gamma: pd.Series) -> pd.Series:
    """
    Get alpha from Corwin-_get_betaSchultz algorithm, (p.285, Snippet 19.1).

    :param high: Series containing high prices
    :rtype high: pd.Series
    :param low: Series containing low prices
    :type low: pd.Series
    :return: Alpha estimates
    :rtype: pd.Series
    """
    den = 3 - 2 * 2 ** 0.5
    alpha = (2 ** 0.5 - 1) * (beta ** 0.5) / den
    alpha -= (gamma / den) ** 0.5
    alpha[alpha < 0] = 0  # Set negative alphas to 0 (see p.727 of paper)
    return alpha


class CorwinSchultzSpread(BaseEstimator, TransformerMixin):
    """
    Corwin Schultz spread estimator.

    See page 284 of Advances in Financial Machine Learning by Marcos Lopez de
    Prado for additional information.

    :param window: The window size
    :type window: int
    :param high_column_name: The name of the high column
    :type high_column_name_get_beta: str, optional
    :param low_column_name: The name of the low column
    :type low_column_name: str, optional
    """

    def __init__(
        self,
        window: int = 20,
        high_column_name: str = static.HIGH,
        low_column_name: str = static.LOW,
    ):
        self.window = window
        self.high_column_name = high_column_name
        self.low_column_name = low_column_name

    def fit(self, X, y=None, **fit_params):
        return self

    def transform(self, X: pd.DataFrame) -> np.ndarray:

        check_missing_columns(X, [self.high_column_name, self.low_column_name])

        beta = _get_beta(X[self.high_column_name], X[self.low_column_name], self.window)
        gamma = _get_gamma(X[self.high_column_name], X[self.low_column_name])
        alpha = _get_alpha(beta, gamma)
        spread = 2 * (np.exp(alpha) - 1) / (1 + np.exp(alpha))

        return spread.values.reshape(-1, 1)


class BeckersParkinsonVolatility(BaseEstimator, TransformerMixin):
    """
    Get Bekker-Parkinson volatility from gamma and beta in Corwin-Schultz algorithm, (p.286, Snippet 19.2).

    See page 284 of Advances in Financial Machine Learning by Marcos Lopez de
    Prado for additional information.

    :param window: The window size
    :type window: int
    :param high_column_name: The name of the high column
    :type high_column_name: str, optional
    :param low_column_name: The name of the low column
    :type low_column_name: str, optional
    """

    def __init__(
        self,
        window: int = 20,
        high_column_name: str = static.HIGH,
        low_column_name: str = static.LOW,
    ):
        self.window = window
        self.high_column_name = high_column_name
        self.low_column_name = low_column_name

    def fit(self, X, y=None, **fit_params):
        return self

    def transform(self, X: pd.DataFrame) -> np.ndarray:

        check_missing_columns(X, [self.high_column_name, self.low_column_name])

        beta = _get_beta(X[self.high_column_name], X[self.low_column_name], self.window)
        gamma = _get_gamma(X[self.high_column_name], X[self.low_column_name])

        k2 = (8 / np.pi) ** 0.5
        den = 3 - 2 * 2 ** 0.5
        sigma = (2 ** -0.5 - 1) * beta ** 0.5 / (k2 * den)
        sigma += (gamma / (k2 ** 2 * den)) ** 0.5
        sigma[sigma < 0] = 0
        return sigma.values.reshape(-1, 1)



import numpy as np
import pandas as pd
from mizarlabs import static
from mizarlabs.transformers.utils import check_missing_columns
from sklearn.base import BaseEstimator
from sklearn.base import TransformerMixin


class VPIN(BaseEstimator, TransformerMixin):
    """
    Implement the Volume-Synchronized Probability of Informed Trading.

    We assume that the index of the dataframe in input is based on the close time

    See page 292 of Advances in Financial Machine Learning by Marcos Lopez de
    Prado for additional information.

    :param window: The window size used for the calculation of the VPIN
    :type window: int
    :param base_asset_volume_column_name: name of the column where the volume (sum of the base asset quantity per trade)
                                          of the time bar is stored, defaults to config.BASE_ASSET_VOLUME
    :type base_asset_volume_column_name: str, optional
    :param base_asset_buy_volume_column_name: name of the column where the volume (sum of the buy base asset quantity
                                              per trade) of the time bar is stored, defaults to
                                              config.BASE_ASSET_BUY_VOLUME
    :type base_asset_buy_volume_column_name: str, optional
    :param base_asset_sell_volume_column_name: name of the column where the volume (sum of the sell base asset quantity
                                               per trade) of the time bar is stored, defaults to
                                               config.BASE_ASSET_SELL_VOLUME
    :type base_asset_sell_volume_column_name: str, optional
    """

    def __init__(
        self,
        window: int = 20,
        base_asset_volume_column_name: str = static.BASE_ASSET_VOLUME,
        base_asset_buy_volume_column_name: str = static.BASE_ASSET_BUY_VOLUME,
        base_asset_sell_volume_column_name: str = static.BASE_ASSET_SELL_VOLUME,
    ):
        self.window = window
        self.base_asset_volume_column_name = base_asset_volume_column_name
        self.base_asset_buy_volume_column_name = base_asset_buy_volume_column_name
        self.base_asset_sell_volume_column_name = base_asset_sell_volume_column_name

    def fit(self, X, y=None, **fit_params):
        return self

    def transform(self, X: pd.DataFrame) -> np.ndarray:
        check_missing_columns(
            X,
            [
                self.base_asset_volume_column_name,
                self.base_asset_buy_volume_column_name,
                self.base_asset_sell_volume_column_name,
            ],
        )

        total_volume_windowed_sum = X.rolling(self.window)[
            self.base_asset_volume_column_name
        ].sum()

        diff_volume = (
            (
                X[self.base_asset_buy_volume_column_name]
                - X[self.base_asset_sell_volume_column_name]
            )
            .abs()
            .rolling(
                self.window,
            )
            .sum()
        )

        vpin_feature = diff_volume / total_volume_windowed_sum

        vpin_feature.iloc[self.window :] = vpin_feature[self.window :].fillna(0)

        assert (
            ~vpin_feature[self.window :].isna().any()
        ), "No NaNs are allowed after the warm-up period of the rolling window"

        return vpin_feature.values.reshape(-1, 1)



"""
Implementation of Trend-Scanning labels described in `Advances in Financial Machine Learning: Lecture 3/10
<https://papers.ssrn.com/sol3/papers.cfm?abstract_id=2708678>`_
"""
from typing import Tuple
from typing import Union

import numpy as np
import pandas as pd
from mizarlabs.static import EVENT_END_TIME
from mizarlabs.static import LABEL
from mizarlabs.static import RETURN
from mizarlabs.static import T_VALUE
from mizarlabs.structural_breaks.sdfc import get_beta_and_beta_var
from numba import jit
from numba import prange
from sklearn.base import BaseEstimator
from sklearn.base import TransformerMixin


class TrendScannerLabeling(BaseEstimator, TransformerMixin):
    """
    `Trend scanning <https://papers.ssrn.com/sol3/papers.cfm?abstract_id=3257419>`_ is both a classification and
    regression labeling technique.
    That can be used in the following ways:
    1. Classification: By taking the sign of t-value for a given observation we can set {-1, 1} labels to define the
       trends as either downward or upward.
    2. Classification: By adding a minimum t-value threshold you can generate {-1, 0, 1} labels for downward, no-trend,
       upward.
    3. The t-values can be used as sample weights in classification problems.
    4. Regression: The t-values can be used in a regression setting to determine the magnitude of the trend.
    The output of this algorithm is a DataFrame with t1 (time stamp for the farthest observation), t-value, returns for
    the trend, and bin.
    :param t_events: filtered events, array/list of pd.Timestamps, defaults to None
    :type t_events: Union[np.ndarray, list, None], optional
    :param look_forward_window: maximum look forward window used to get the trend value, defaults to 20
    :type look_forward_window: int, optional
    :param min_sample_length: minimum sample length used to fit regression, defaults to 5
    :type min_sample_length: int, optional
    :param step: optimal t-value index is searched every 'step' indices, defaults to 1
    :type step: int, optional
    """

    def __init__(
        self,
        t_events: Union[np.ndarray, list, None] = None,
        look_forward_window: int = 20,
        min_sample_length: int = 5,
        step: int = 1,
    ):
        assert min_sample_length < look_forward_window, (
            f"look_forward_window is {look_forward_window} "
            f"and should be larger than min_sample_length {min_sample_length}"
        )
        self.t_events = t_events
        self.look_forward_window = look_forward_window
        self.min_sample_length = min_sample_length
        self.step = step

    def fit(self, y):
        return self

    def _check_input(self, input_series: pd.Series):
        assert isinstance(
            input_series, pd.Series
        ), "Please provide only a Pandas series with close prices with datetime indices."
        if self.t_events is not None:
            assert set(input_series.index.values).issuperset(set(self.t_events)), (
                f"The following time indices are in t_events but "
                f"not in the provided series "
                f"{set(input_series.index.values) - set(self.t_events)}"
            )

    def _check_output(self, output_df: pd.DataFrame):
        required_columns = [EVENT_END_TIME, T_VALUE, RETURN, LABEL]

        assert set(required_columns) == set(
            output_df.columns
        ), f"Missing {set(required_columns) - set(output_df.columns)} in output DataFrame."
        assert set(self.t_events) == set(
            output_df.index.values
        ), f"Missing following time indices in output DataFrame: {set(self.t_events) - set(output_df.index.values)}"

    def transform(self, y: pd.Series) -> pd.DataFrame:
        """Scans for trends in the provided series and provides
        a DataFrame with results.

        DataFrame contains the start_time, event_end_time, t_value,
        return and label.

        :param y: series used to label the data set
        :type y: pd.Series
        :return: DataFrame with as index the start time and
                 in the columns the event_end_time, t_value, return and
                 label.
        :rtype: pd.DataFrame
        """
        self._check_input(y)
        df = self._transform(y)
        self._check_output(df)
        return df

    def _transform(self, y: pd.Series) -> pd.DataFrame:
        """Scans for trends in the provided series and provides
        a DataFrame with results.

        DataFrame contains the start_time, event_end_time, t_value,
        return and label.

        :param y: series used to label the data set
        :type y: pd.Series
        :return: DataFrame with as index the start time and
                 in the columns the event_end_time, t_value, return and
                 label.
        :rtype: pd.DataFrame
        """

        # check if only subset of indices is looped over
        # else scan every index
        if self.t_events is None:
            self.t_events = y.index.values

        # get t values and end indices per index
        t_values_array, end_indices_array = _trend_scan_each_index(
            y_array=y.values.astype(np.float64),
            indices=y.index.get_indexer(self.t_events),
            min_sample_length=self.min_sample_length,
            look_forward_window=self.look_forward_window,
            step=self.step,
        )

        # convert end indices into datetime
        event_end_times_array = [
            y.index[int(i)] if not np.isnan(i) else i for i in end_indices_array
        ]

        # cast results into dataframe and compute the returns and the sign of the return
        labels = pd.DataFrame(
            {EVENT_END_TIME: event_end_times_array, T_VALUE: t_values_array},
            index=self.t_events,
        )
        non_nan_labels = labels[EVENT_END_TIME].dropna()
        labels.loc[non_nan_labels.index, RETURN] = (
            y.loc[non_nan_labels].values / y.loc[non_nan_labels.index].values - 1
        )
        labels[LABEL] = np.sign(labels.t_value)
        return labels


@jit(parallel=True, nopython=True)
def _trend_scan_each_index(
    y_array: np.ndarray,
    indices: np.ndarray,
    min_sample_length: int,
    look_forward_window: int,
    step: int,
) -> Tuple[np.ndarray, np.ndarray]:
    """Scans for the trend in the index.

    :param look_forward_window: maximum look forward window used to get the trend value, defaults to 20
    :type look_forward_window: int, optional
    :param min_sample_length: minimum sample length used to fit regression, defaults to 5
    :type min_sample_length: int, optional
    :param step: optimal t-value index is searched every 'step' indices, defaults to 1
    :type step: int, optional

    :param y_array: series to trend scan
    :type y_array: np.ndarray
    :param indices: indices in the series to scan
    :type indices: np.ndarray
    :param min_sample_length: minimum sample length used to fit regression
    :type min_sample_length: int
    :param look_forward_window: maximum look forward window used to get the trend value
    :type look_forward_window: int
    :param step: optimal t-value index is searched every 'step' indices
    :type step: int
    :return: Tuple with the max t-values and their respective location in the series
    :rtype: Tuple[np.ndarray, np.ndarray]
    """
    # array to save the end time index values
    end_indices_array = np.empty(len(indices), dtype=np.float64)
    # array to save the t values per index
    t_values_array = np.empty(len(indices), dtype=np.float64)

    # iterate over each index in t_events
    for i in prange(len(indices)):
        index = indices[i]

        # select a subset of the series of interest based on forward looking window
        y_subset = y_array[index : (index + look_forward_window)].astype(np.float64)

        # check if sufficient data else set values to NaN
        if y_subset.shape[0] == look_forward_window:

            # compute max t value and its index location in the subset of y
            (
                t_values_array[i],
                index_max_t_value,
            ) = _get_t_val_and_event_end_time_index_at_index(
                y_subset, min_sample_length, step
            )

            # convert index location to index in whole of y
            end_indices_array[i] = index + index_max_t_value + min_sample_length - 1

        else:
            end_indices_array[i] = np.nan
            t_values_array[i] = np.nan
    return t_values_array, end_indices_array


@jit(nopython=True)
def _get_t_val_and_event_end_time_index_at_index(
    y_subset: np.ndarray,
    min_sample_length: int,
    step: int,
) -> Tuple[np.float64, np.float64]:
    """Returns the maximum t-value and its index location.

    Loop over possible look-ahead windows to get the one
    which yields maximum t values for b_1 regression coef

    :param y_subset: series being trend scanned
    :type y_subset: np.ndarray
    :param min_sample_length: minimum sample length used to fit regression
    :type min_sample_length: int
    :param step: optimal t-value index is searched every 'step' indices
    :type step: int
    :return: tuple with max t-value and its index location
    :rtype: Tuple[np.float64, np.float64]
    """

    # init array to stores t values in
    t_values_array = np.empty(y_subset.shape[0] - min_sample_length)

    # expand forward looking window, compute and store t value, keeping into account
    # the min_sample_length and the stepsize
    for forward_window in np.arange(min_sample_length, y_subset.shape[0], step):
        # y{t}:y_{t+l}
        y_subset_forward_window = np.ascontiguousarray(
            y_subset[:forward_window].reshape(-1, 1)
        )

        # Array of [1, 0], [1, 1], [1, 2], ... [1, l] # b_0, b_1 coefficients
        X_constant = np.ones_like(y_subset_forward_window, dtype=np.float64)
        X_trend = (
            np.arange(y_subset_forward_window.shape[0])
            .astype(np.float64)
            .reshape(-1, 1)
        )
        X_subset_forward_window = np.ascontiguousarray(
            np.concatenate((X_constant, X_trend), axis=1)
        )

        # get OLS estimates
        b_mean_, b_var_ = get_beta_and_beta_var(
            X_subset_forward_window, y_subset_forward_window
        )

        # compute t value for trend coefficient
        t_values_array[forward_window - min_sample_length] = b_mean_[1, 0] / np.sqrt(
            b_var_[1, 1]
        )

    # identify max abs t-value and its location
    index_max_t_value = np.argmax(np.abs(t_values_array))
    max_t_value = t_values_array[index_max_t_value]
    return max_t_value, index_max_t_value



import itertools
import numbers
from abc import ABCMeta
from abc import abstractmethod
from typing import Tuple
from typing import Union
from warnings import warn

import numpy as np
import pandas as pd
from mizarlabs.transformers.targets.labeling import EVENT_END_TIME
from scipy import sparse
from sklearn.base import BaseEstimator
from sklearn.base import ClassifierMixin
from sklearn.ensemble import BaggingClassifier
from sklearn.ensemble._bagging import BaseBagging
from sklearn.ensemble._base import _partition_estimators
from sklearn.metrics import accuracy_score
from sklearn.tree import DecisionTreeClassifier
from sklearn.utils import check_array
from sklearn.utils import check_consistent_length
from sklearn.utils import check_random_state
from sklearn.utils import check_X_y
from sklearn.utils import indices_to_mask
from sklearn.utils._joblib import delayed
from sklearn.utils._joblib import Parallel
from sklearn.utils.random import sample_without_replacement
from sklearn.utils.validation import has_fit_parameter

from .bootstrapping import get_ind_matrix
from .bootstrapping import seq_bootstrap

MAX_INT = np.iinfo(np.int32).max


def _generate_random_features(
    random_state: np.random.RandomState,
    bootstrap: bool,
    n_population: int,
    n_samples: int,
) -> np.array:
    """
    Draw randomly sampled indices.

    :param random_state: Random state objecc
    :type random_state: np.random.RandomState
    :param bootstrap: Boolean indicating whether drawing with or without replacement.
    :type bootstrap: bool
    :param n_population: The size of the set to sample from.
    :type n_population: int
    :param n_samples: The number of samples to draw.
    :type n_samples: int
    :return: array indicating which samples have been drawn.
    :rtype: np.array
    """
    # Draw sample indices
    if bootstrap:
        indices = random_state.randint(0, n_population, n_samples)
    else:
        indices = sample_without_replacement(
            n_population, n_samples, random_state=random_state
        )

    return indices


def _generate_bagging_indices(
    random_state: np.random.RandomState,
    bootstrap_features: bool,
    n_features: int,
    max_features: int,
    max_samples: int,
    ind_mat: np.array,
    update_probs_every: int,
) -> Tuple[np.array, np.array]:
    """
    Randomly draw feature and sample indices.

    :param random_state: Random state object.
    :type random_state: np.random.RandomState
    :param bootstrap_features: Boolean indicating whether features
                               need to be bootstrapped.
    :type bootstrap_features: bool
    :param n_features: Number of features to draw.
    :type n_features: int
    :param max_features: Max number of features available.
    :type max_features: int
    :param max_samples: Max number of samples to draw.
    :type max_samples: int
    :param ind_mat: Indicator matrix from triple barrier events
    :type ind_mat: np.array
    :param update_probs_every: only update the sampling probabilities with average uniqueness after
                               update_probs_every times, this will speed up training, but at the cost that you
                               do not sample perfectly according to the average uniqueness
    :return: tuple with two arrays indicating
             the features and samples drawn respectively.
    :rtype: Tuple[np.aray, np.array]
    """
    # Get valid random state
    random_state = check_random_state(random_state)

    # Draw indices
    feature_indices = _generate_random_features(
        random_state, bootstrap_features, n_features, max_features
    )
    sample_indices = seq_bootstrap(
        ind_mat, sample_length=max_samples, random_state=random_state, update_probs_every=update_probs_every,
    )

    return feature_indices, sample_indices


def _parallel_build_estimators(
    n_estimators: int,
    ensemble: BaseBagging,
    X: pd.DataFrame,
    y: pd.Series,
    ind_mat: sparse.csc_matrix,
    sample_weight: pd.Series,
    seeds: np.array,
    total_n_estimators: int,
    verbose: int,
    update_probs_every: int,
) -> tuple:
    """
    Private function used to build a batch of estimators within a job.

    :param n_estimators: Number of base estimators to build in batch.
    :type n_estimators: int
    :param ensemble: Ensemble model.
    :type ensemble: BaseBagging
    :param X: DataFrame with features.
    :type X: pd.DataFrame
    :param y: Series with labels.
    :type y: pd.Series
    :param ind_mat: indicator matrix from triple barrier events
    :type ind_mat: sparse.csc_matrix
    :param sample_weight: Series with sample weights.
    :type sample_weight: pd.Series
    :param seeds: Array with seeds for random state.
    :type seeds: np.array
    :param total_n_estimators: Total number of estimators in ensemble.
    :type total_n_estimators: int
    :param verbose: Indicating how much to print.
    :type verbose: int
    :param update_probs_every: only update the sampling probabilities with average uniqueness after
                               update_probs_every times, this will speed up training, but at the cost that you
                               do not sample perfectly according to the average uniqueness
    :return: Tuple with estimators, features and the estimator indices
    :rtype: tuple
    """
    # Retrieve settings
    n_samples, n_features = X.shape
    max_features = ensemble._max_features
    max_samples = ensemble._max_samples
    bootstrap_features = ensemble.bootstrap_features
    support_sample_weight = has_fit_parameter(ensemble.base_estimator_, "sample_weight")

    if not support_sample_weight and sample_weight is not None:
        raise ValueError("The base estimator doesn't support sample weight")

    # Build estimators
    estimators = []
    estimators_features = []
    estimators_indices = []

    for i in range(n_estimators):
        if verbose > 1:
            print(
                "Building estimator %d of %d for this parallel run "
                "(total %d)..." % (i + 1, n_estimators, total_n_estimators)
            )

        random_state = np.random.RandomState(seeds[i])
        estimator = ensemble._make_estimator(append=False, random_state=random_state)

        # Draw random feature, sample indices
        features, indices = _generate_bagging_indices(
            random_state,
            bootstrap_features,
            n_features,
            max_features,
            max_samples,
            ind_mat,
            update_probs_every,
        )

        # Draw samples, using sample weights, and then fit
        if support_sample_weight:
            if sample_weight is None:
                curr_sample_weight = np.ones((n_samples,))
            else:
                curr_sample_weight = sample_weight.copy()

            sample_counts = np.bincount(indices, minlength=n_samples)
            curr_sample_weight *= sample_counts

            estimator.fit(X[:, features], y, sample_weight=curr_sample_weight)

        else:
            estimator.fit((X[indices])[:, features], y[indices])

        estimators.append(estimator)
        estimators_features.append(features)
        estimators_indices.append(indices)

    return estimators, estimators_features, estimators_indices


class SequentiallyBootstrappedBaseBagging(BaseBagging, metaclass=ABCMeta):
    """
    Base class for Sequentially Bootstrapped Classifier and Regressor, extension of sklearn's BaseBagging
    """

    @abstractmethod
    def __init__(
        self,
        samples_info_sets: pd.Series,
        price_bars: pd.DataFrame,
        base_estimator: BaseEstimator = None,
        n_estimators: int = 10,
        max_samples: Union[int, float] = 1.0,
        max_features: Union[int, float] = 1.0,
        bootstrap_features: bool = False,
        oob_score: bool = False,
        warm_start: bool = False,
        n_jobs: int = None,
        random_state: Union[int, np.random.RandomState, None] = None,
        verbose: int = 0,
        event_end_time_column_name: str = EVENT_END_TIME,
        update_probs_every: int = 1,
    ):
        super().__init__(
            base_estimator=base_estimator,
            n_estimators=n_estimators,
            bootstrap=True,
            max_samples=max_samples,
            max_features=max_features,
            bootstrap_features=bootstrap_features,
            oob_score=oob_score,
            warm_start=warm_start,
            n_jobs=n_jobs,
            random_state=random_state,
            verbose=verbose,
        )

        self.event_end_time_column_name = event_end_time_column_name
        self.samples_info_sets = samples_info_sets
        self.price_bars = price_bars
        self._ind_mat = None
        self.update_probs_every = update_probs_every

        # Used for create get ind_matrix subsample during cross-validation
        self.timestamp_int_index_mapping = pd.Series(
            index=samples_info_sets.index, data=range(self.ind_mat.shape[1])
        )

        self.X_time_index = None  # Timestamp index of X_train

    @property
    def ind_mat(self):
        if self._ind_mat is None:
            self._ind_mat = get_ind_matrix(
                self.samples_info_sets, self.price_bars, self.event_end_time_column_name
            )
        return self._ind_mat

    def fit(
        self,
        X: pd.DataFrame,
        y: pd.Series,
        sample_weight: pd.Series = None,
    ):
        """Build a Sequentially Bootstrapped Bagging ensemble of estimators from the training
           set (X, y).
        Parameters
        ----------
        X : {array-like, sparse matrix} of shape = [n_samples, n_features]
            The training input samples. Sparse matrices are accepted only if
            they are supported by the base estimator.
        y : array-like, shape = [n_samples]
            The target values (class labels in classification, real numbers in
            regression).
        sample_weight : array-like, shape = [n_samples] or None
            Sample weights. If None, then samples are equally weighted.
            Note that this is supported only if the base estimator supports
            sample weighting.
        Returns
        -------
        self : object
        """
        return self._fit(X, y, self.max_samples, sample_weight=sample_weight)

    def _fit(
        self,
        X: pd.DataFrame,
        y: pd.Series,
        max_samples: Union[int, float] = None,
        max_depth: int = None,
        sample_weight: pd.Series = None,
    ):
        """Build a Sequentially Bootstrapped Bagging ensemble of estimators from the training
           set (X, y).
        Parameters
        ----------
        X : {array-like, sparse matrix} of shape = [n_samples, n_features]
            The training input samples. Sparse matrices are accepted only if
            they are supported by the base estimator.
        y : array-like, shape = [n_samples]
            The target values (class labels in classification, real numbers in
            regression).
        max_samples : int or float, optional (default=None)
            Argument to use instead of self.max_samples.
        max_depth : int, optional (default=None)expiritation
            Override value used when constructing base estimator. Only
            supported if the base estimator has a max_depth parameter.
        sample_weight : array-like, shape = [n_samples] or None
            Sample weights. If None, then samples are equally weighted.
            Note that this is supported only if the base estimator supports
            sample weighting.
        Returns
        -------
        self : object
        """

        assert isinstance(X, pd.DataFrame), "X should be a dataframe with time indices"
        assert isinstance(y, pd.Series), "y should be a series with time indices"
        assert isinstance(
            X.index, pd.DatetimeIndex
        ), "X index should be a DatetimeIndex"
        assert isinstance(
            y.index, pd.DatetimeIndex
        ), "y index should be a DatetimeIndex"

        random_state = check_random_state(self.random_state)
        self.X_time_index = X.index  # Remember X index for future sampling

        assert set(self.timestamp_int_index_mapping.index).issuperset(
            set(self.X_time_index)
        ), "The ind matrix timestamps should have all the timestamps in the training data"

        # Generate subsample ind_matrix (we need this during subsampling cross_validation)
        ind_mat = self.ind_mat.tocsc()

        subsampled_ind_mat = ind_mat[
            :, self.timestamp_int_index_mapping.loc[self.X_time_index]
        ]

        # Convert data (X is required to be 2d and indexable)
        X, y = check_X_y(
            X, y, ["csr", "csc"], dtype=None, force_all_finite=False, multi_output=True
        )
        if sample_weight is not None:
            sample_weight = check_array(sample_weight, ensure_2d=False)
            check_consistent_length(y, sample_weight)

        # Remap output
        n_samples, self.n_features_ = X.shape
        self._n_samples = n_samples
        y = self._validate_y(y)

        # Check parameters
        self._validate_estimator()

        # Validate max_samples
        if not isinstance(max_samples, (numbers.Integral, np.integer)):
            max_samples = int(max_samples * X.shape[0])

        if not (0 < max_samples <= X.shape[0]):
            raise ValueError("max_samples must be in (0, n_samples]")

        # Store validated integer row sampling value
        self._max_samples = max_samples

        # Validate max_features
        if isinstance(self.max_features, (numbers.Integral, np.integer)):
            max_features = self.max_features
        elif isinstance(self.max_features, np.float):
            max_features = self.max_features * self.n_features_
        else:
            raise ValueError("max_features must be int or float")

        if not (0 < max_features <= self.n_features_):
            raise ValueError("max_features must be in (0, n_features]")

        max_features = max(1, int(max_features))

        # Store validated integer feature sampling value
        self._max_features = max_features

        if self.warm_start and self.oob_score:
            raise ValueError("Out of bag estimate only available if warm_start=False")

        if not self.warm_start or not hasattr(self, "estimators_"):
            # Free allocated memory, if anyexpiritation
            self.estimators_ = []
            self.estimators_features_ = []
            self.sequentially_bootstrapped_samples_ = []

        n_more_estimators = self.n_estimators - len(self.estimators_)

        if n_more_estimators < 0:
            raise ValueError(
                "n_estimators=%d must be larger or equal to "
                "len(estimators_)=%d when warm_start==True"
                % (self.n_estimators, len(self.estimators_))
            )

        elif n_more_estimators == 0:
            warn(
                "Warm-start fitting without increasing n_estimators does not "
                "fit new trees."
            )
            return self

        # Parallel loop
        n_jobs, n_estimators, starts = _partition_estimators(
            n_more_estimators, self.n_jobs
        )
        total_n_estimators = sum(n_estimators)

        # Advance random state to state after training
        # the first n_estimators
        if self.warm_start and len(self.estimators_) > 0:
            random_state.randint(MAX_INT, size=len(self.estimators_))

        seeds = random_state.randint(MAX_INT, size=n_more_estimators)
        self._seeds = seeds

        all_results = Parallel(n_jobs=n_jobs, verbose=self.verbose,)(
            delayed(_parallel_build_estimators)(
                n_estimators[i],
                self,
                X,
                y,
                subsampled_ind_mat,
                sample_weight,
                seeds[starts[i] : starts[i + 1]],
                total_n_estimators,
                verbose=self.verbose,
                update_probs_every=self.update_probs_every,
            )
            for i in range(n_jobs)
        )

        # Reduce
        self.estimators_ += list(
            itertools.chain.from_iterable(t[0] for t in all_results)
        )
        self.estimators_features_ += list(
            itertools.chain.from_iterable(t[1] for t in all_results)
        )
        self.sequentially_bootstrapped_samples_ += list(
            itertools.chain.from_iterable(t[2] for t in all_results)
        )

        if self.oob_score:
            self._set_oob_score(X, y)

        self._ind_mat = None

        return self


class SequentiallyBootstrappedBaggingClassifier(
    SequentiallyBootstrappedBaseBagging, BaggingClassifier, ClassifierMixin
):
    """
    A Sequentially Bootstrapped Bagging classifier is an ensemble meta-estimator that fits base
    classifiers each on random subsets of the original dataset generated using
    Sequential Bootstrapping sampling procedure and then aggregate their individual predictions (
    either by voting or by averaging) to form a final prediction. Such a meta-estimator can typically be used as
    a way to reduce the variance of a black-box estimator (e.g., a decision
    tree), by introducing randomization into its construction procedure and
    then making an ensemble out of it.
    :param samples_info_sets: pd.Series, The information range on which each record is constructed from
        *samples_info_sets.index*: Time when the information extraction started.
        *samples_info_sets.value*: Time when the information extraction ended.
    :param price_bars: pd.DataFrame
        Price bars used in samples_info_sets generation
    :param base_estimator: object or None, optional (default=None)
        The base estimator to fit on random subsets of the dataset.
        If None, then the base estimator is a decision tree.
    :param n_estimators: int, optional (default=10)
        The number of base estimators in the ensemble.
    :param max_samples: int or float, optional (default=1.0)
        The number of samples to draw from X to train each base estimator.
        If int, then draw `max_samples` samples. If float, then draw `max_samples * X.shape[0]` samples.
    :param max_features: int or float, optional (default=1.0)
        The number of features to draw from X to train each base estimator.
        If int, then draw `max_features` features. If float, then draw `max_features * X.shape[1]` features.
    :param bootstrap_features: boolean, optional (default=False)
        Whether features are drawn with replacement.
    :param oob_score: bool, optional (default=False)
        Whether to use out-of-bag samples to estimate
        the generalization error.
    :param warm_start: bool, optional (default=False)
        When set to True, reuse the solution of the previous call to fit
        and add more estimators to the ensemble, otherwise, just fit
        a whole new ensemble.
    :param n_jobs: int or None, optional (default=None)
        The number of jobs to run in parallel for both `fit` and `predict`.
        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
        ``-1`` means using all processors.
    :param random_state: int, RandomState instance or None, optional (default=None)
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by `np.random`.
    :param verbose: int, optional (default=0)
        Controls the verbosity when fitting and predicting.
    :param event_end_time_column_name: str, optional (default=EXPIRATION_BARRIER)
        name of the column with the expiration barrier dates.
    :param update_probs_every: int, optional (default=1)
        Only update the sampling probabilities with average uniqueness after
        update_probs_every times, this will speed up training, but at the cost that you
        do not sample perfectly according to the average uniqueness
    :ivar base_estimator_: estimator
        The base estimator from which the ensemble is grown.
    :ivar estimators_: list of estimators
        The collection of fitted base estimators.
    :ivar estimators_samples_: list of arrays
        The subset of drawn samples (i.e., the in-bag samples) for each base
        estimator. Each subset is defined by an array of the indices selected.
    :ivar estimators_features_: list of arrays
        The subset of drawn features for each base estimator.
    :ivar classes_: array of shape = [n_classes]
        The classes labels.
    :ivar n_classes_: int or list
        The number of classes.
    :ivar oob_score_: float
        Score of the training dataset obtained using an out-of-bag estimate.
    :ivar oob_decision_function_: array of shape = [n_samples, n_classes]
        Decision function computed with out-of-bag estimate on the training
        set. If n_estimators is small it might be possible that a data point
        was never left out during the bootstrap. In this case,
        `oob_decision_function_` might contain NaN.
    """

    def __init__(
        self,
        samples_info_sets: pd.Series,
        price_bars: pd.DataFrame,
        base_estimator: BaseEstimator = None,
        n_estimators: int = 10,
        max_samples: Union[int, float] = 1.0,
        max_features: Union[int, float] = 1.0,
        bootstrap_features: bool = False,
        oob_score: bool = False,
        warm_start: bool = False,
        n_jobs: int = None,
        random_state: Union[int, np.random.RandomState, None] = None,
        verbose: int = 0,
        event_end_time_column_name: str = EVENT_END_TIME,
        update_probs_every: int = 1,
    ):
        super().__init__(
            samples_info_sets=samples_info_sets,
            price_bars=price_bars,
            base_estimator=base_estimator,
            n_estimators=n_estimators,
            max_samples=max_samples,
            max_features=max_features,
            bootstrap_features=bootstrap_features,
            oob_score=oob_score,
            warm_start=warm_start,
            n_jobs=n_jobs,
            random_state=random_state,
            verbose=verbose,
            event_end_time_column_name=event_end_time_column_name,
            update_probs_every=update_probs_every,
        )

    def _validate_estimator(self):
        """
        Check the estimator and set the base_estimator_ attribute.
        """
        super(BaggingClassifier, self)._validate_estimator(
            default=DecisionTreeClassifier()
        )

    def _set_oob_score(self, X: pd.DataFrame, y: pd.Series):
        """
        Calculates and sets the out of bag score.

        :param X: DataFrame with features.
        :type X: pd.DataFrame
        :param y: Series with labels.
        :type y: pd.Series
        """
        n_samples = y.shape[0]
        n_classes_ = self.n_classes_

        predictions = np.zeros((n_samples, n_classes_))

        for estimator, samples, features in zip(
            self.estimators_,
            self.sequentially_bootstrapped_samples_,
            self.estimators_features_,
        ):
            # Create mask for OOB samples
            mask = ~indices_to_mask(samples, n_samples)

            if hasattr(estimator, "predict_proba"):
                predictions[mask, :] += estimator.predict_proba(
                    (X[mask, :])[:, features]
                )

            else:
                p = estimator.predict((X[mask, :])[:, features])
                j = 0

                for i in range(n_samples):
                    if mask[i]:
                        predictions[i, p[j]] += 1
                        j += 1

        if (predictions.sum(axis=1) == 0).any():
            warn(
                "Some inputs do not have OOB scores. "
                "This probably means too few estimators were used "
                "to compute any reliable oob estimates."
            )

        oob_decision_function = predictions / predictions.sum(axis=1)[:, np.newaxis]
        oob_score = accuracy_score(y, np.argmax(predictions, axis=1))

        self.oob_decision_function_ = oob_decision_function
        self.oob_score_ = oob_score



"""
Static variables that we do not expect to change
"""
# Tick data column properties
TIMESTAMP_UNIT = "ms"

# Time bars column names
OPEN = "open"
LOW = "low"
HIGH = "high"
CLOSE = "close"
BASE_ASSET_VOLUME = "base_asset_volume"
BASE_ASSET_BUY_VOLUME = "base_asset_buy_volume"
BASE_ASSET_SELL_VOLUME = "base_asset_sell_volume"
QUOTE_ASSET_VOLUME = "quote_asset_volume"

FIRST_TIMESTAMP = "first_timestamp"
TIMESTAMP = "time"  # will be in ms

# Column names for trading
SIZE = "size"
SIDE = "side"
NUMBER_EXPIRATION_BARS_COLUMN = "num_expiration_bars"
EVENT_END_TIME = "event_end_time"
DAILY_VOL = "daily_vol"
STOP_LOSS = "stop_loss"
PROFIT_TAKING = "profit_taking"
RETURN = "return"
LABEL = "label"
T_VALUE = "t_value"



import numpy as np
import pandas as pd
from mizarlabs.model.bootstrapping import get_ind_matrix
from mizarlabs.static import CLOSE
from mizarlabs.transformers.sampling.average_uniqueness import AverageUniqueness
from mizarlabs.static import EVENT_END_TIME
from mizarlabs.transformers.utils import check_missing_columns
from mizarlabs.transformers.utils import convert_to_timestamp
from sklearn.base import BaseEstimator
from sklearn.base import TransformerMixin


class SampleWeightsByReturns(BaseEstimator, TransformerMixin):
    """
    Calculate the sample weights by absolute returns.

    See page 69 of Advances in Financial Machine Learning by Marcos Lopez de
    Prado for additional information.

    :param event_end_time_column_name: The column name of the event end time
    :type event_end_time_column_name: str, optional
    :param close_column_name: The column name of the close price
    :type close_column_name: str, optional
    """

    def __init__(
        self,
        event_end_time_column_name: str = EVENT_END_TIME,
        close_column_name: str = CLOSE,
    ):

        self._event_end_time_column_name = event_end_time_column_name
        self._close_column_name = close_column_name

    def transform(self, X: pd.DataFrame) -> pd.Series:
        check_missing_columns(
            X, [self._close_column_name, self._event_end_time_column_name]
        )

        indicators_matrix, ind_mat_indices = get_ind_matrix(
            samples_info_sets=X[self._event_end_time_column_name],
            price_bars=X,
            event_end_time_column_name=self._event_end_time_column_name,
            return_indices=True,
        )

        num_concurrent_events = pd.Series(
            indicators_matrix.tocsc().sum(axis=1).A1, index=ind_mat_indices
        ).loc[X.index]

        returns = np.log(X[self._close_column_name]).diff()

        weights_array = self._calculate_weights(
            convert_to_timestamp(X.index.values),
            convert_to_timestamp(X[self._event_end_time_column_name].values),
            num_concurrent_events.values,
            returns.values,
        )

        weights = pd.Series(weights_array, index=X.index)

        return weights

    @staticmethod
    def _calculate_weights(
        bars_index: np.ndarray,
        expiration_barriers: np.ndarray,
        num_concurrent_events: np.ndarray,
        returns: np.ndarray,
    ) -> np.ndarray:
        """
        Calculate the weights by absolute returns

        It is based on numpy arrays for optimization

        :param bars_index:
        :type bars_index: np.ndarray
        :param expiration_barriers:
        :type expiration_barriers: np.ndarray
        :param num_concurrent_events:
        :type num_concurrent_events: np.ndarray
        :param returns:
        :type returns: np.ndarray
        :return: The weights by absolute returns
        :rtype: np.ndarray
        """
        # init weights array
        weights = np.zeros_like(bars_index, dtype=np.float64)
        for i in range(len(expiration_barriers)):
            # creating mask between start time and expiration barrier
            mask = np.greater_equal(bars_index, bars_index[i]) * np.less_equal(
                bars_index, expiration_barriers[i]
            )
            # calculating weights based on the returns and the concurrent events
            weights[i] = np.abs(np.nansum(returns[mask] / num_concurrent_events[mask]))

        return weights


class SampleWeightsByTimeDecay(AverageUniqueness):
    """
    Calculate the sample weights by time decay.

    See page 70 of Advances in Financial Machine Learning by Marcos Lopez de
    Prado for additional information.

    :param event_end_time_column_name: The column name of the event end time
    :type event_end_time_column_name: str, optional
    :param minimum_decay_weight: Is the minimum desired value in the decay weights
        - minimum_decay_weight = 1 means there is no time decay
        - 0 < minimum_decay_weight < 1 means that weights decay linearly over
          time, but every observation still receives a strictly positive weight, regadless of how old
        - minimum_decay_weight = 0 means that weights converge linearly to zero, as they become older
        - minimum_decay_weight < 0 means that the oldest portion of the observations
          receive zero weight (i.e they are erased from memory)
    :type minimum_decay_weight: float
    """

    def __init__(
        self,
        minimum_decay_weight: float,
        event_end_time_column_name: str = EVENT_END_TIME,
    ):
        super().__init__(event_end_time_column_name)
        self._decay_factor = minimum_decay_weight

    def transform(self, X: pd.DataFrame) -> pd.Series:
        assert self._decay_factor <= 1, "Decay factor should be less or equal to 1"
        check_missing_columns(X, [self._event_end_time_column_name])
        av_uniqueness = super().transform(X)
        time_decay_weights = av_uniqueness.sort_index().cumsum()
        if self._decay_factor >= 0:
            slope = (1 - self._decay_factor) / time_decay_weights.iloc[-1]
        else:
            slope = 1 / ((self._decay_factor + 1) * time_decay_weights.iloc[-1])
        const = 1 - slope * time_decay_weights.iloc[-1]
        time_decay_weights = const + slope * time_decay_weights
        time_decay_weights[time_decay_weights < 0] = 0  # Weights can't be negative

        return time_decay_weights



from sklearn.base import BaseEstimator
from sklearn.base import TransformerMixin
import pandas as pd


class LocalMaxToCloseRatio(BaseEstimator, TransformerMixin):
    def __init__(self, window):
        self.window = window

    def transform(self, X: pd.DataFrame) -> pd.Series:
        return (X.high.rolling(window=self.window).max() / X.close).values.reshape(
            -1, 1
        )



from sklearn.base import BaseEstimator
from sklearn.base import TransformerMixin
import pandas as pd
from mizarlabs.transformers.technical.factory import TAFactory
import numpy as np


class MeanReversionStrategy:
    classes_ = [-1.0, 0.0, 1.0]
    n_classes_ = 3

    def __init__(self, threshold: float, short_time_period: int):
        self.threshold = threshold
        self.short_time_period = short_time_period
        self.short_ema_transformer = TAFactory().create_transformer(
            "EMA", "close", kw_args={"timeperiod": short_time_period}
        )
        self.fitted_ = True

    def predict(self, X: pd.DataFrame) -> np.ndarray:
        short_ema = pd.Series(self.short_ema_transformer.transform(X).flatten(), index=X.index)

        diff_close_short_ema = (X["close"] - short_ema) / short_ema
        signal = pd.Series(0, index=X.index)
        signal.loc[(diff_close_short_ema > self.threshold)] = -1
        signal.loc[(diff_close_short_ema < -self.threshold)] = 1
        return signal.values

    def predict_proba(self, X: pd.DataFrame) -> np.ndarray:
        predictions = self.predict(X)
        probabilities = np.zeros(shape=(len(predictions), len(self.classes_)))
        for i, class_value in enumerate(self.classes_):
            probabilities[:, i] = predictions == class_value
        return probabilities



from mizarlabs.transformers.technical.factory import TAFactory
import pandas as pd
import numpy as np
from sklearn.base import BaseEstimator
from sklearn.base import TransformerMixin


class MACDHistogramCrossOver(BaseEstimator, TransformerMixin):
    def __init__(
        self,
        fast_period: int = 12,
        slow_period: int = 26,
        signal_period: int = 9,
        column_name: str = "close",
    ):

        if fast_period > slow_period:
            raise ValueError("Fast should be smaller than slow")
        self.fast_period_ = fast_period
        self.slow_period_ = slow_period
        self.signal_period_ = signal_period
        self.column_name = column_name
        self.transformer_ = TAFactory().create_transformer(
            "MACD",
            self.column_name,
            kw_args={
                "fastperiod": fast_period,
                "slowperiod": slow_period,
                "signalperiod": signal_period,
            },
        )

    def fit(self, X, y=None, **fit_params):
        return self

    def transform(self, X: pd.DataFrame):

        macd_df = pd.DataFrame(self.transformer_.transform(X))

        return np.sign(macd_df[2]).diff().values.reshape(-1, 1) / 2



import pandas as pd
import numpy as np
from sklearn.base import BaseEstimator
from sklearn.base import TransformerMixin
from mizarlabs import static


class RollingZScoreTransformer(BaseEstimator, TransformerMixin):
    """Transformer which calculates rolling z score based on a transformation
    of a dataframe into a series.

    :param window: lookbackwindow of rolling z score
    :type window: int
    :param to_series_func: function to apply on dataframe to transform into a series
    :type to_series_func: callable
    """
    def __init__(self, window: int, to_series_func: callable):
        self.window = window
        self.to_series_func = to_series_func

    def transform(self, X: pd.DataFrame) -> pd.Series:
        """Transform dataframe into rolling z score.

        :param X: DataFrame to transform
        :type X: pd.DataFrame
        :return: Series with rolling z scores
        :rtype: pd.Series
        """
        series = self.to_series_func(X)
        rolling_z_score = self.calc_z_score(series)
        return rolling_z_score.values.reshape(-1, 1)

    def calc_z_score(self, X: pd.Series) -> pd.Series:
        """calculated the rolling z score

        :param X: series to compute z score of
        :type X: pd.Series
        :return: series with rolling z scores
        :rtype: pd.Series
        """
        X_rolling_mean = X.rolling(window=self.window).mean()
        X_rolling_std = X.rolling(window=self.window).std()
        rolling_z_score = (X - X_rolling_mean) / X_rolling_std
        return rolling_z_score


def bar_arrival_time(bars_df: pd.DataFrame) -> pd.Series:
    return bars_df.index.to_series().astype(int).diff()

def buy_sell_diff(bars_df: pd.DataFrame) -> pd.Series:
    return bars_df.quote_asset_buy_volume - bars_df.quote_asset_sell_volume

def average_buy_size(bars_df: pd.DataFrame) -> pd.Series:
    return bars_df.quote_asset_buy_volume / bars_df.num_buy_ticks

def high_to_low_ratio(bars_df: pd.DataFrame) -> pd.Series:
    return (bars_df.high / bars_df.low)

class RollingZScoreTransformerFactory:
    """Factory for creating predefined rolling z score transformers.
    """
    types = dict(
        bar_arrival_time=bar_arrival_time,
        buy_sell_diff=buy_sell_diff,
        average_buy_size=average_buy_size,
        high_to_low_ratio=high_to_low_ratio,
    )

    @classmethod
    def get_transformer(self, key: str, window: int) -> RollingZScoreTransformer:
        """Builds a rolling z score transformer.

        Select from the following: bar_arrival_time, buy_sell_diff, average_buy_size, high_to_low_ratio.

        :param key: name of the transformer to build
        :type key: str
        :param window: lookback window for rolling mean and standard deviation
        :type window: int
        :raises KeyError: cannot build transformers, which are not registered.
        :return: Transformer with requested specification
        :rtype: RollingZScoreTransformer
        """
        if key not in self.types:
            raise KeyError(f"{key} does not exist, please select from {', '.join(self.types.keys())}")
        return RollingZScoreTransformer(window=window, to_series_func=self.types[key])




import re
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple
from typing import Union

import attr
import numpy as np
import pandas as pd
import talib
from sklearn.preprocessing import FunctionTransformer


@attr.s
class TAFactory:
    """
    Factory that creates sklearn transformers for function available in the TA-lib package.
    """

    ta_func_dict = talib.__dict__
    _real = "real"

    @staticmethod
    def get_function_groups() -> Dict[str, str]:
        """
        Returns a dictionary with TA-lib function names specified per group.

        :return: the key-value pairs are group names by list of function names
        :rtype: Dict[str, str]
        """
        return talib.get_function_groups()

    @staticmethod
    def get_functions() -> List[str]:
        """
        Returns all the available function names in TA-lib package.

        :return: list with function names in the TA-lib package.
        :rtype: List[str]
        """
        return talib.get_functions()

    def get_args_kwargs(self, func_name: str) -> Tuple[List[str]]:
        """
        Returns a tuple with a list of valid argument names and keyword argument names for the TA function.

        :param func_name: function name in the TA-lib package
        :type func_name: str
        :return: first element is the list with valid argument names and the second element is the list with
                 valid keyword argument names.
        :rtype: Tuple[List[str]]
        """
        ta_func = self.get_ta_func(func_name)
        signature = re.search(r"\((.+)\)", ta_func.__doc__).group(1)

        args_kwargs_names = signature.split("[, ")
        if len(args_kwargs_names) == 1:
            args_names, kwargs_names = args_kwargs_names[0], None
        else:
            args_names, kwargs_names = args_kwargs_names

        args_names_list = args_names.split(", ")

        if kwargs_names is not None:
            kwargs_names_list = kwargs_names.strip("=?]").split("=?, ")
        else:
            kwargs_names_list = None
        return args_names_list, kwargs_names_list

    def get_ta_func(self, func_name: str) -> callable:
        """
        Returns the TA function callable.

        :param func_name: name of the function in TA-lib package
        :type func_name: str
        :raises ValueError: if provided name does not exist as a function in the TA-lib package an error is raised
        :return: the TA function which creates TA features
        :rtype: callable
        """
        if func_name not in self.ta_func_dict.keys():
            raise ValueError(
                f"The func_name {func_name} is not available. Please check if the name is spelled correctly."
            )
        ta_func = self.ta_func_dict.get(func_name)
        return ta_func

    def create_transformer(
        self,
        func_name: str,
        col_to_transform: Union[List[str], str] = None,
        col_to_transform_mapping: Dict[str, str] = None,
        kw_args: dict = None,
    ) -> FunctionTransformer:
        """
        Creates a transformer for a given function available in the TA-lib package.

        :param func_name: name of function in TA-lib package
        :type func_name: str
        :param col_to_transform: name of the column in dataframe to transform, defaults to None
        :type col_to_transform: Union[List[str], str], optional
        :param col_to_transform_mapping: remapping of names in dataframe to valid arg names in the TA function,
                                         defaults to None
        :type col_to_transform_mapping: Dict[str, str], optional
        :param kw_args: kwargs for the TA function, defaults to None
        :type kw_args: dict, optional
        :raises TypeError: transformer only works on pandas DataFrames
        :raises ValueError: transformer expect certain columns to be available in the dataframe to transform
        :return: scikit-learn FunctionTransformer version of the TA function
        :rtype: FunctionTransformer
        """

        ta_func = self.get_ta_func(func_name)

        args_names_list, kwargs_names_list = self.get_args_kwargs(func_name)

        args_names_list = self._check_and_set_args(
            func_name, args_names_list, col_to_transform, col_to_transform_mapping
        )

        kw_args = self._check_and_set_kwargs(func_name, kw_args, kwargs_names_list)

        def ta_func_transformed(X: pd.DataFrame) -> np.ndarray:
            if not isinstance(X, pd.DataFrame):
                raise TypeError(
                    f"X must be a pandas DataFrame, but got type {type(X)}."
                )
            if not set(X.columns).issuperset(set(args_names_list)):
                raise ValueError(
                    f"Expected the following columns in the dataframe: {set(args_names_list) - set(X.columns)}"
                )
            args = [X[i] for i in args_names_list]
            output = ta_func(*args, **kw_args)

            # parse output to 2-d arrays to enable concatenating
            if isinstance(output, tuple):
                output = pd.concat(output, axis=1).values

            if output.ndim != 2:
                if isinstance(output, pd.Series):
                    output = output.values
                output = output.reshape(-1, 1)

            return output

        return FunctionTransformer(ta_func_transformed)

    def _check_and_set_args(
        self,
        func_name: str,
        args_names_list: List[str],
        col_to_transform: Union[str, None],
        col_to_transform_mapping: Union[str, None],
    ) -> List[str]:
        """
        Checks and sets the args for the TA function.

        :param func_name: name of the TA function
        :type func_name: str
        :param args_names_list: list with valid argument names
        :type args_names_list: List[str]
        :param col_to_transform: if the TA function takes in an array this will specify which column to transform in
                                 the dataframe
        :type col_to_transform: Union[str, None]
        :param col_to_transform_mapping: maps valid argument names to column names in the dataframe to transform
        :type col_to_transform_mapping: Union[str, None]
        :raises ValueError: column name to transform is not provided, i.e. transformer does not know which column
                            in the dataframe to transform
        :raises ValueError: transformer takes in two arrays, so both column need to be specified
        :raises TypeError: the column mapping is not a dictionary
        :raises ValueError: the mapping contains names that are not valid argument names
        :return: list with valid column names to extract from the dataframe to transform
        :rtype: List[str]
        """
        # ta func takes in an array and transforms it directly
        # since X is a DataFrame with multiple columns, it must be
        # specified which column is being transformed
        if self._real in args_names_list:
            if not isinstance(col_to_transform, str):
                raise ValueError(
                    f"The function {func_name} requires col_to_transform to be specified."
                )
            idx_real = args_names_list.index(self._real)
            args_names_list[idx_real] = col_to_transform

        # functions that take in two arrays need to be specified as a list with two column names
        if all(f"real{i}" in args_names_list for i in range(2)):
            if not (
                isinstance(col_to_transform, list)
                and all(isinstance(i, str) for i in col_to_transform)
            ):
                raise ValueError(
                    f"Selected function {func_name} requires two columns to be "
                    "specified to apply the transformation on, please provide a list with column names. "
                )
            for i in range(2):
                idx_real = args_names_list.index(f"real{i}")
                args_names_list[idx_real] = col_to_transform[i]

        # some expected columns can have different names in the dataframe
        # this part will map the expected name in the ta func to the name in the dataframe
        if col_to_transform_mapping is not None:
            if not isinstance(col_to_transform_mapping, dict):
                raise TypeError(
                    f"Expected a dictionary for col_to_transform_mapping, but got {type(col_to_transform_mapping)}."
                )
            if set(col_to_transform_mapping.keys()).issubset(set(args_names_list)):
                for i, arg_name_i in enumerate(args_names_list):
                    if arg_name_i in col_to_transform_mapping.keys():
                        args_names_list[i] = col_to_transform_mapping[arg_name_i]
            else:
                raise ValueError(
                    f"The function {func_name} does not accept "
                    f"{set(col_to_transform_mapping.keys()) - set(args_names_list)} as arguments. "
                    f"The only valid arguments are {args_names_list}."
                )
        return args_names_list

    @staticmethod
    def _check_and_set_kwargs(
        func_name: str,
        kw_args: Union[None, Dict[str, any]],
        kwargs_names_list: List[str],
    ) -> Dict[str, Any]:
        """
        Checks if the provided kwargs for the TA function are valid.

        :param func_name: name of the TA function
        :type func_name: str
        :param kw_args: dictionary with kwargs that are being set for the TA function
        :type kw_args: Union[None, Dict[str, any]]
        :param kwargs_names_list: list with valid kwarg names
        :type kwargs_names_list: List[str]
        :raises error: if the TA function has no kwargs or kwargs do not exist errors are raised
        :raises ValueError: TA function has no kwargs, but kwargs are being set.
        :raises ValueError: TA function does not have all the kwargs that are being set.
        :return: Dictionary with valid kwargs for the TA function.
        :rtype: Dict[str, Any]
        """
        # raise error if kwargs are being set, while there are no kwargs or some do not exist
        if kwargs_names_list is None and kw_args is not None:
            raise ValueError(
                f"Selected function {func_name} has no key word arguments, but got {kw_args}."
            )
        if kwargs_names_list is not None and kw_args is not None:
            if not set(kw_args.keys()).issubset(set(kwargs_names_list)):
                raise ValueError(
                    f"The kwargs {set(kw_args.keys()) - set(kwargs_names_list)} are not valid kwargs "
                    f"for the function {func_name}. Valid kwargs are: {kwargs_names_list}."
                )

        if kw_args is None:
            kw_args = dict()

        return kw_args



import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator
from sklearn.base import TransformerMixin


class MovingAverageCrossOver(BaseEstimator, TransformerMixin):
    """
    Moving average crossover transformer

    It transform the input data to a 1 when the the fast moving average goes
    above the slow moving average and -1 when the slow moving average goes
    above the fast moving average. When fill_between_crossovers is true, then
    the transfromation will be equal to 1 when the fast moving average is
    above the slow one and -1 viceversa.

    The transformer is specified by the fast and slow moving average
    number of bars.

    :param fast: number of bars to use for the moving average of the fast
                 moving average
    :type fast: int
    :param slow: number of bars to use for the moving average of the slow
                  moving average
    :type slow: int
    :param column_name: The name of the column in input that will be used for the
                        transformation.
    :type column_name: str
    :param fill_between_crossovers: whether or not to fill the crossover value untill
                                    the next crossover.
    :type fill_between_crossovers: bool
    """

    def __init__(
        self,
        fast: int,
        slow: int,
        column_name: str,
        fill_between_crossovers: bool = False,
    ):

        if fast > slow:
            raise ValueError("Fast should be smaller than slow")
        self.fast_ = fast
        self.slow_ = slow
        self.column_name = column_name
        self.fill_between_crossovers = fill_between_crossovers

    def fit(self, X, y=None, **fit_params):
        return self

    def transform(self, X: pd.DataFrame):
        fast_moving_average = X[self.column_name].rolling(window=self.fast_).mean()
        slow_moving_average = X[self.column_name].rolling(window=self.slow_).mean()

        # filling the zeros with nan so that we can forward fill the nans
        # with the previous crossover value
        if self.fill_between_crossovers:
            side = pd.Series(np.nan, index=X.index)
        else:
            side = pd.Series(0, index=X.index)

        side.loc[
            self._get_down_cross_indices(
                fast_ma=fast_moving_average, slow_ma=slow_moving_average
            )
        ] = -1
        side.loc[
            self._get_up_cross_indices(
                fast_ma=fast_moving_average, slow_ma=slow_moving_average
            )
        ] = 1

        # forward filling nans with previous crossover values
        side.ffill(inplace=True)
        # filling nans at the top of the series with zeros
        side.fillna(0, inplace=True)

        return side.values.reshape(-1, 1)

    @staticmethod
    def _get_up_cross_indices(
        fast_ma: pd.Series, slow_ma: pd.Series
    ) -> pd.DatetimeIndex:
        # previous value is smaller or equal
        cond_0 = fast_ma.shift(1) <= slow_ma.shift(1)
        # current value is greater
        cond_1 = fast_ma > slow_ma
        return fast_ma[(cond_0) & (cond_1)].index

    @staticmethod
    def _get_down_cross_indices(fast_ma: pd.Series, slow_ma: pd.Series) -> pd.Series:
        # previous value is greater or equal
        cond_0 = fast_ma.shift(1) >= slow_ma.shift(1)
        # current value is smaller
        cond_1 = fast_ma < slow_ma
        return fast_ma[(cond_0) & (cond_1)].index


class MovingAverageCrossOverPredictor(MovingAverageCrossOver):
    classes_ = [-1.0, 0.0, 1.0]
    n_classes_ = 3

    def predict(self, X: pd.DataFrame):
        return self.transform(X).flatten()

    def predict_proba(self, X: pd.DataFrame):
        predictions = self.predict(X)
        probabilities = np.zeros(shape=(len(predictions), len(self.classes_)))
        for i, class_value in enumerate(self.classes_):
            probabilities[:, i] = predictions == class_value
        return probabilities


class ExponentialWeightedMovingAverageDifference:
    def __init__(
        self,
        fast: int,
        slow: int,
        column_name: str,
        normalised: bool = True,
    ):

        if fast > slow:
            raise ValueError("Fast should be smaller than slow")
        self.fast_ = fast
        self.slow_ = slow
        self.normalised = normalised
        self.column_name = column_name

    def fit(self, X, y=None, **fit_params):
        return self

    def transform(self, X: pd.DataFrame) -> np.ndarray:
        fast_ewma = X[self.column_name].ewm(span=self.fast_).mean()
        slow_ewma = X[self.column_name].ewm(span=self.slow_).mean()

        ewma_difference = fast_ewma - slow_ewma

        if self.normalised:
            return (ewma_difference / (fast_ewma + 1)).values.reshape(
                -1, 1
            )  # plus one to avoid zero division error
        else:
            return ewma_difference.values.reshape(-1, 1)



import logging
from typing import Dict
from typing import Tuple
from typing import Union

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator
from sklearn.base import TransformerMixin
from sklearn.pipeline import FeatureUnion
from sklearn.pipeline import Pipeline
from sklearn.utils.validation import check_is_fitted
from sklearn.utils.validation import check_X_y

from mizarlabs.static import CLOSE
from mizarlabs.static import NUMBER_EXPIRATION_BARS_COLUMN
from mizarlabs.static import PROFIT_TAKING
from mizarlabs.static import SIDE
from mizarlabs.static import SIZE
from mizarlabs.static import STOP_LOSS
from mizarlabs.transformers.targets.labeling import get_daily_vol
from mizarlabs.transformers.trading.bet_sizing import BetSizingFromProbabilities
from mizarlabs.transformers.utils import check_missing_columns


class MizarFeatureUnion(FeatureUnion):
    def fit_transform(self, X, y=None, **fit_params):
        Xs = super().fit_transform(X, y=None, **fit_params)

        return pd.DataFrame(Xs, index=X.index)

    def transform(self, X):
        Xs = super().transform(X)

        return pd.DataFrame(Xs, index=X.index)


class MizarPipeline(Pipeline):
    """
    Implementation of pipeline that allows sample_weight as a fit argument
    """

    def fit(self, X, y, sample_weight=None, **fit_params):
        if sample_weight is not None:
            fit_params[self.steps[-1][0] + "__sample_weight"] = sample_weight
        return super().fit(X, y, **fit_params)


class StrategySignalPipeline:
    """
    A trading strategy.

    A trading strategy can include machine learning models or simple technical
    indicator transformers. From their outputs the strategy decides whether or
    not to take a position and its size.

    This strategy must have a primary model. A metalabeling model and
    a bet sizer are optional.

    The simplest setting includes only a primary model. In this case the side
    is calculated with the predict of the primary model, while the size is
    calculated from the predict_proba of the primary model.

    Adding a bet sizer means that the size is calculated from the bet sizer
    and not anymore from the primary model probabilities. The bet sizer
    calculates the bet size from the probabilities of the primary model
    predictions.

    When metalabeling model is set then the size comes from the metalabeling
    model, unless a bet sizer is set and in this case the bet sizer calculates
    the size from the probabilites provided by the metalabeling model

    :param feature_transformers_primary_model: The feature transformer that
                                               transforms the data for the
                                               primary model
    :type feature_transformers_primary_model: TransformerMixin
    :param feature_transformers_metalabeling_model: The feature transformer that
                                                    transforms the data for the
                                                    metalabeling model
    :type feature_transformers_metalabeling_model: TransformerMixin

    :param metalabeling_use_proba_primary_model: Whether to use probabilities
                                                 of the primary model as
                                                 features in the metalabeling
                                                 model
    :type metalabeling_use_proba_primary_model: bool
    :param metalabeling_use_predictions_primary_model: Whether to use
                                                       predictions of the
                                                       primary model as feature
                                                       in the metalabeling
                                                       model
    :param bet_sizer: The transformer to use for the calculation of the bet size
    :type bet_sizer: BetSizingFromProbabilities
    """

    _primary_model_features = "primary_model_features"
    _primary_model_predictions = "primary_model_predictions"
    _primary_model_proba = "primary_model_proba"
    _metalabeling_model_features = "metalabeling_model_features"
    _metalabeling_model_predictions = "metalabeling_model_predictions"
    _metalabeling_model_proba = "metalabeling_model_proba"

    _align_methods = {
        "mean": np.mean,
        "min": np.min,
        "max": np.max,
    }

    def __init__(
        self,
        feature_transformers_primary_model: Union[
            Dict[str, Union[TransformerMixin, None]]
        ],
        align_on: str,
        align_how: Dict[str, str],
        feature_transformers_metalabeling_model: Union[
            Dict[str, Union[TransformerMixin, None]]
        ] = None,
        metalabeling_use_proba_primary_model: bool = True,
        metalabeling_use_predictions_primary_model: bool = True,
        bet_sizer: Union[BetSizingFromProbabilities, None] = None,
    ):
        self.primary_model = None
        self.metalabeling_model = None

        self.feature_transformers_primary_model = (
            feature_transformers_primary_model
            if feature_transformers_primary_model
            else {}
        )

        self.align_on_ = align_on
        self.align_how_ = align_how

        self.feature_transformers_metalabeling_model = (
            feature_transformers_metalabeling_model
            if feature_transformers_metalabeling_model
            else {}
        )

        self.metalabeling_use_proba_primary_model = metalabeling_use_proba_primary_model
        self.metalabeling_use_predictions_primary_model = (
            metalabeling_use_predictions_primary_model
        )
        self.bet_sizer = bet_sizer

    def set_primary_model(self, primary_model: BaseEstimator):
        self.primary_model = primary_model
        return self

    def set_metalabeling_model(self, metalabeling_model: BaseEstimator):
        self.metalabeling_model = metalabeling_model
        return self

    def _align_on(
        self, X_features_dict: Dict[str, Dict[str, pd.DataFrame]]
    ) -> Dict[str, Dict[str, pd.DataFrame]]:
        """
        Aligns the dataframes with features with the prespecified
        leading dataframe with features.

        :param X_features_dict: dictionary with features for the primary (and
                                metalabeling model if set) with keys
                                "primary_model_features" and
                                "metalabeling_model_features".
        :type X_features_dict: Dict[str, Dict[str, pd.DataFrame]]
        :return: dictionary with aligned features for the primary (and
                 metalabeling model if set) with keys "primary_model_features"
                 and "metalabeling_model_features".
        :rtype: Dict[str, Dict[str, pd.DataFrame]]
        """
        if not self.align_on_:
            return X_features_dict

        alignment_index = X_features_dict[self._primary_model_features][
            self.align_on_
        ].index

        for model_features_key in [
            self._primary_model_features,
            self._metalabeling_model_features,
        ]:
            if model_features_key is None:
                continue

            for feature_key in X_features_dict[model_features_key].keys():
                if feature_key != self.align_on_:

                    feature_df = X_features_dict[model_features_key][feature_key]

                    # -np.inf equivalent of datetime
                    previous_index = pd.to_datetime("1-1-1700", utc=feature_df.index.tz)

                    aligned_feature_df = pd.DataFrame(
                        columns=feature_df.columns, index=alignment_index
                    )

                    diff_days_start = (feature_df.index[0] - alignment_index[0]).days

                    if (
                        abs(diff_days_start)
                        > (alignment_index[-1] - alignment_index[0]).days / 10
                    ):
                        earlier_later = "later" if diff_days_start > 0 else "earlier"

                        logging.warning(
                            f"The start date of the features dataset"
                            f" {feature_key} starts "
                            f"{(alignment_index[-1] - alignment_index[0]).days} "
                            f"days {earlier_later}"
                            f" than the alignment data"
                        )

                    for index in alignment_index:
                        feature_df_subset = feature_df.loc[
                            (feature_df.index > previous_index)
                            & (feature_df.index <= index)
                        ]
                        if feature_df_subset.empty:
                            continue

                        elif feature_df_subset.shape[0] == 1:
                            aligned_feature_df.loc[index] = feature_df_subset.values

                        else:
                            aligned_feature_df.loc[index] = self._align_methods[
                                self.align_how_[feature_key]
                            ](feature_df_subset.values, axis=0)
                        previous_index = index

                    X_features_dict[model_features_key][
                        feature_key
                    ] = aligned_feature_df.ffill()

        return X_features_dict

    def _get_pred_proba_features(
        self,
        primary_model: BaseEstimator,
        X_pred_features: pd.DataFrame,
        X_val: pd.DataFrame,
        y_val: pd.Series,
    ) -> pd.DataFrame:
        """
        Get the probabilities of the primary model predictions to be used in
        the metalabeling model as features

        :param primary_model: The primary model object
        :type primary_model: BaseEstimator
        :param X_pred_features: DataFrame with predictions from the primary model,
                                which will be used as features for the metalabeling
                                model.
        :type X_pred_features: pd.DataFrame
        :param X_val: Features of the primary model used for validation during
                      cross validation.
        :type X_val: pd.DataFrame
        :param y_val: Class labels of the primary used for validation during
                      cross validation.
        :type y_val: pd.Series
        :return: DataFrame with predictions and probabilities from the primary
                 model to be used as features in the metalabeling model.
        :rtype: pd.DataFrame
        """
        if not hasattr(primary_model, "predict_proba"):
            raise AttributeError("Primary model does not have a predict_proba method.")

        y_pred_proba_val = primary_model.predict_proba(X_val)
        pred_proba_cols = [
            f"primary_side_pred_proba_{i}" for i in range(y_pred_proba_val.shape[1])
        ]

        if not set(pred_proba_cols).issubset(set(X_pred_features.columns)):
            X_pred_proba = pd.DataFrame(
                y_pred_proba_val,
                index=y_val.index,
                columns=[
                    f"primary_side_pred_proba_{i}"
                    for i in range(y_pred_proba_val.shape[1])
                ],
            )
            X_pred_features = pd.concat([X_pred_features, X_pred_proba], axis=1)
        else:
            X_pred_features.loc[y_val.index, pred_proba_cols] = y_pred_proba_val

        return X_pred_features

    @staticmethod
    def _check_y_metalabel(y_metalabel: pd.Series):
        """
        Check whether metalabels can be used for training the metalabeling
        model.

        :param y_metalabel: Metalabels calculated from the primary model
        :type y_metalabel: pd.Series
        :return:
        """
        unique_metalabels = np.unique(y_metalabel)
        if len(unique_metalabels) == 1:
            sentence_string = "incorrect" if 0.0 in unique_metalabels else "correct"
            raise ValueError(
                f"The primary model is always {sentence_string} "
                f"and so it is not possible to train the "
                f"metalabeling model"
            )

    @staticmethod
    def _align_X_dict_and_y(
        X_dict: Dict[str, pd.DataFrame], y: pd.Series, drop_na: bool = True
    ) -> Tuple[pd.DataFrame, pd.Series]:
        """
        Align the features in X and the target in y

        :param X_dict: A dictionary containing features for the primary and
                       metalabeling model
        :type X_dict: Dict[str, pd.DataFrame]
        :param y: The target to which the features need to be aligned with
        :type y: pd.Series

        :return: Features, target and sample weights aligned
        :rtype: Tuple[pd.DataFrame, pd.Series, pd.Series]
        """

        if drop_na:
            y_no_nans = y.dropna()
            X = pd.concat([df.loc[y_no_nans.index] for df in X_dict.values()], axis=1)
            X_aligned = X.dropna()
            y_aligned = y.loc[X_aligned.index]
            check_X_y(X_aligned, y_aligned)
            return X_aligned, y_aligned
        else:
            X = pd.concat([df for df in X_dict.values()], axis=1)
            return X, y

    def transform(
        self,
        X_dict: Dict[str, pd.DataFrame],
    ) -> Dict[str, Dict[str, pd.DataFrame]]:
        """
        Runs the feature transformers (if available) on the data.

        :param X_dict: A dictionary containing features for the primary and
                       metalabeling model
        :type X_dict: Dict[str, pd.DataFrame]
        :return: Dictionary containing the features per each model transformed
        :rtype: Dict[str, Dict[str, pd.DataFrame]]
        """
        expected_keys = {
            *self.feature_transformers_primary_model.keys(),
            *self.feature_transformers_metalabeling_model.keys(),
        }
        assert expected_keys == set(X_dict.keys()), (
            f"The keys in X_dict should be exactly the same as the name "
            f"of the feature generators of the primary (and metalabeling) model. "
            f"Expected {expected_keys}, but got {set(X_dict.keys())}"
        )
        return {
            self._primary_model_features: {
                name_transformer: transformer.transform(X_dict[name_transformer])
                if transformer
                else X_dict[name_transformer]
                for name_transformer, transformer in self.feature_transformers_primary_model.items()
            },
            self._metalabeling_model_features: {
                name_transformer: transformer.transform(X_dict[name_transformer])
                if transformer
                else X_dict[name_transformer]
                for name_transformer, transformer in self.feature_transformers_metalabeling_model.items()
            },
        }

    def _get_features_for_metalabeling(
        self,
        X_primary_aligned: pd.DataFrame,
        X_features_dict: Dict[str, Dict[str, pd.DataFrame]],
    ) -> pd.DataFrame:
        """
        Create the aligned dataset for the metalabeling model adding the
        predictions and probabilities from the primary model

        :param X_primary_aligned: The dateset aligned for the primary model
        :type X_primary_aligned: pd.DataFrame
        :param X_features_dict: Dict containing the transformed features for the
                                primary and metalabeling model
        :type X_features_dict: Dict[str, Dict[str, pd.DataFrame]]
        :return: Aligned dataset for the metalabeling model
        :rtype: pd.DataFrame
        """
        X_metalabel_aligned = pd.concat(
            [df for df in X_features_dict[self._metalabeling_model_features].values()],
            axis=1,
        )
        X_metalabel_aligned.dropna(inplace=True)
        # add primary model predict values for metalabel model as features
        if self.metalabeling_use_predictions_primary_model:
            X_metalabel_aligned["primary_side_pred"] = pd.Series(
                self.primary_model.predict(X_primary_aligned),
                index=X_primary_aligned.index,
            )
        # add primary model predict proba values for metalabel model as features
        if self.metalabeling_use_proba_primary_model:
            X_primary_proba_features = pd.DataFrame(
                self.primary_model.predict_proba(X_primary_aligned),
                index=X_primary_aligned.index,
                columns=[
                    f"primary_side_pred_proba_{i}"
                    for i in range(self.primary_model.n_classes_)
                ],
            )
            # to avoid the creation of unwanted nans the join between the the
            # metalabeling aligned and the primary model predictions is inner.
            # Without using an inner join nan values will be created when
            # the indices of the two dataframe are not the same
            X_metalabel_aligned = pd.concat(
                [X_metalabel_aligned, X_primary_proba_features], axis=1, join="inner"
            )

        if X_metalabel_aligned.isna().any().any():
            raise ValueError(
                "Nan values in the metalabeling aligned dataset "
                "are not allowed. Please check your metalabeling "
                "model does not create any nan values when "
                "it predicts"
            )
        return X_metalabel_aligned

    def predict(
        self,
        X_dict: Dict[str, pd.DataFrame],
    ) -> Dict[str, pd.Series]:
        """
        Predict the classes for the primary and metalabeling model

        :param X_dict: A dictionary containing features for the primary and
                       meta-labeling model
        :type X_dict: Dict[str, pd.DataFrame]
        :return: Predicted probabilities for primary and metalabeling model
        :rtype: Dict[str, pd.DataFrame]
        """
        if self.metalabeling_model:
            check_is_fitted(self.metalabeling_model)

        target_index = X_dict[self.align_on_].index

        for df_name, df in X_dict.items():
            if target_index[0] > df.index[-1]:
                raise ValueError(
                    f"There is no intersection between {df_name} and {self.align_on_}"
                )

        X_features_dict = self.transform(X_dict)
        X_features_dict = self._align_on(X_features_dict)
        X_primary_aligned = pd.concat(
            [df for df in X_features_dict[self._primary_model_features].values()],
            axis=1,
        )
        X_primary_aligned.dropna(inplace=True)

        # get predictions primary model
        pred_primary = pd.Series(
            self.primary_model.predict(X_primary_aligned), index=X_primary_aligned.index
        )
        predictions = {
            self._primary_model_predictions: pred_primary,
            self._metalabeling_model_predictions: None,
        }

        if self.metalabeling_model:
            X_metalabel_aligned = self._get_features_for_metalabeling(
                X_primary_aligned, X_features_dict
            )
            pred_metalabeling = pd.Series(
                self.metalabeling_model.predict(X_metalabel_aligned),
                index=X_metalabel_aligned.index,
            )
            predictions[self._metalabeling_model_predictions] = pred_metalabeling

        return predictions

    def predict_proba(self, X_dict: Dict[str, pd.DataFrame]) -> Dict[str, pd.DataFrame]:
        """
        Predict the probabilities for the primary and metalabeling model

        :param X_dict: A dictionary containing features for the primary and
                       meta-labeling model
        :type X_dict: Dict[str, pd.DataFrame]
        :return: Predicted probabilities for primary and metalabeling model
        :rtype: Dict[str, pd.DataFrame]
        """

        # transforming and aligning X_dict
        X_features_dict = self.transform(X_dict)
        X_features_dict = self._align_on(X_features_dict)

        # concatenating all the features dataframes in
        # one single dataframe that can be used for
        # predictions
        X_primary_aligned = pd.concat(
            [df for df in X_features_dict[self._primary_model_features].values()],
            axis=1,
        )
        X_primary_aligned.dropna(inplace=True)

        # get probabilities from primary model
        prob_primary_df = pd.DataFrame(
            self.primary_model.predict_proba(X_primary_aligned),
            index=X_primary_aligned.index,
            columns=self.primary_model.classes_,
        )

        # creating the probabilities dictionary
        # with the probabilites from the primary model
        probabilities = {
            self._primary_model_proba: prob_primary_df,
            self._metalabeling_model_proba: None,
        }

        # if the metalabeling model is present
        # then the metalabeling probabilities are
        # added to the probabilities dictionary
        if self.metalabeling_model:
            check_is_fitted(self.metalabeling_model)

            # aligning metalabeling model's features
            X_metalabel_aligned = self._get_features_for_metalabeling(
                X_primary_aligned, X_features_dict
            )

            # get probabilities from metalabeling model
            pred_metalabeling = pd.DataFrame(
                self.metalabeling_model.predict_proba(X_metalabel_aligned),
                index=X_metalabel_aligned.index,
                columns=self.metalabeling_model.classes_,
            )
            # adding the probabilities to the dictionary
            probabilities[self._metalabeling_model_proba] = pred_metalabeling

        return probabilities

    def get_side_and_size(self, X_dict: Dict[str, pd.DataFrame]) -> pd.DataFrame:
        """
        Calculate the side and size of the position

        :param X_dict: A dictionary containing features for the primary and
                       meta-labeling model
        :type X_dict: Dict[str, pd.DataFrame]
        :return: The sizes of the positions
        :rtype: pd.Series
        """

        side = self.predict(X_dict=X_dict)[self._primary_model_predictions]

        predicted_proba_dict = self.predict_proba(X_dict=X_dict)
        predicted_proba = predicted_proba_dict[
            self._metalabeling_model_proba
            if self.metalabeling_model
            else self._primary_model_proba
        ]

        # TODO: how to implement average_active option in bet_sizer
        # requires event_end_time_column_name
        if self.bet_sizer:
            pred_and_proba = pd.DataFrame(
                data={
                    "prob": predicted_proba.max(axis=1).values,
                    "pred": np.array(
                        [
                            predicted_proba.columns[i]
                            for i in np.argmax(predicted_proba.values, axis=1)
                        ]
                    ),
                },
                index=predicted_proba.index,
            )
            if self.metalabeling_model:
                predicted_side = self.predict(X_dict=X_dict)[
                    self._primary_model_predictions
                ]
                pred_and_proba["side"] = predicted_side
            size = self.bet_sizer.transform(pred_and_proba)

        else:
            if self.metalabeling_model:
                size = predicted_proba[1.0]
            else:
                size = predicted_proba.max(axis=1)

        return pd.DataFrame({"side": side, "size": size}).dropna()

    def create_dataset_metalabeling(
        self,
        X_dict: Dict[str, pd.DataFrame],
        y: pd.Series,
    ) -> Tuple[pd.DataFrame, pd.Series]:

        """
        Produce data set for metalabeling model fitting.

        The primary model is expected to be already set in the pipeline

        :param X_dict: Dictionary containing all the features for the data
                       for the primary and metalabeling model. The data can be
                       bar and/or tick data
        :type X_dict:  Dict[str, pd.DataFrame]
        :param y: Series with class labels for the primary model
        :type y: pd.Series
        :return: The strategy signal pipeline
        :rtype: StrategySignalPipeline
        """
        assert set(np.unique(y.dropna())).issubset({-1, 0, 1})

        # transforming the X_dict using the specified transformers
        X_features_dict = self.transform(X_dict)

        # Aligning the dictionaries to the data from which the
        # target has been constructed
        X_features_dict = self._align_on(X_features_dict)

        # align X_dict a dict consisting of dataframes with features and y
        (X_primary_aligned, y_primary_aligned,) = self._align_X_dict_and_y(
            X_features_dict[self._primary_model_features],
            y,
        )

        X_metalabel = self._get_features_for_metalabeling(
            X_primary_aligned, X_features_dict
        )
        y_metalabel = (
            self.primary_model.predict(X_primary_aligned) == y_primary_aligned
        ).astype(float)

        return (
            X_metalabel.loc[X_metalabel.index.intersection(y_metalabel.index)],
            y_metalabel.loc[X_metalabel.index.intersection(y_metalabel.index)],
        )

    def create_dataset_primary(
        self,
        X_dict: Dict[str, pd.DataFrame],
        y: pd.Series,
        drop_na: bool = True,
    ) -> Tuple[pd.DataFrame, pd.Series]:
        """
        Produce data set for primary model fitting

        :param X_dict: Dictionary containing all the features for the data
                       for the primary and metalabeling model. The data can be
                       bar and/or tick data
        :type X_dict:  Dict[str, pd.DataFrame]
        :param y: Series with class labels for the primary model
        :type y: pd.Series
        :param drop_na:
        :type drop_na: bool
        :return: The strategy signal pipeline
        :rtype: StrategySignalPipeline
        """
        assert set(np.unique(y.dropna())).issubset({-1, 0, 1})

        # transforming the X_dict using the specified transformers
        X_features_dict = self.transform(X_dict)

        # Aligning the dictionaries to the data from which the
        # target has been constructed
        X_features_dict = self._align_on(X_features_dict)

        # align X_dict a dict consisting of dataframes with features and y
        (X_aligned, y_aligned,) = self._align_X_dict_and_y(
            X_features_dict[self._primary_model_features], y, drop_na
        )

        return X_aligned, y_aligned


class StrategyTrader:
    """
    What is my purpose?

    Interacts with the data provider,
    use the strategy pipeline to make a prediction,
    Based on a prediction produces all the information to create a position
    (side, size, expiration, profit taking, stop loss)
    """

    def __init__(
        self,
        strategy_pipeline: StrategySignalPipeline,
        # min number of bars needed to create a prediction
        # (take max of the min of primary and meta-labeling model)
        min_num_bars: int,
        num_expiration_bars: int,  # when does trade expire
        stop_loss_factor: Union[float, None] = None,
        profit_taking_factor: Union[float, None] = None,
        volatility_window: int = 100,
    ):

        self.strategy_pipeline = strategy_pipeline
        self.min_num_bars = min_num_bars
        self.num_expiration_bars = num_expiration_bars
        self.volatility_window = volatility_window
        self.stop_loss_factor = stop_loss_factor
        self.profit_taking_factor = profit_taking_factor
        self.strategy_valid = None
        # TODO: think about a triple barrier labeling with no
        #  volatility adjustment -> check labeling.py

    def _check_X_dict(self, X_dict: Dict[str, pd.DataFrame]):
        """

        :param X_dict: A dictionary containing features for the primary and
                       meta-labeling model
        :type X_dict: Dict[str, pd.DataFrame]
        """
        for df_key, df in X_dict.items():
            if not df.index.is_unique:
                raise ValueError(f"The dataframe {df_key} has duplicated indices")

    def _check_signal_has_valid_output(self, signal_df: pd.DataFrame) -> None:
        """
        Check whether the signal has valid values

        :param signal_df: dataframe containing the size and side of the signal
        :type signal_df: pd.DataFrame
        :return: None
        :rtype: None
        """
        # checking if output of strategy is what we expect
        size_unexpected_values = (
            not (0 <= signal_df[SIZE]).all() or not (signal_df[SIZE] <= 1).all()
        )
        side_unexpected_values = (
            not (-1 <= signal_df[SIDE]).all() or not (signal_df[SIDE] <= 1).all()
        )

        if size_unexpected_values or side_unexpected_values:
            self.strategy_valid = False
        else:
            self.strategy_valid = True

    def create_signal(self, X_dict: Dict[str, pd.DataFrame]) -> pd.DataFrame:
        """
        Create the signal info dataframe (size and side)

        :param X_dict: A dictionary containing features for the primary and
                       meta-labeling model
        :type X_dict: Dict[str, pd.DataFrame]
        :return: dataframe with size and side
        :rtype: pd.DataFrame
        """
        self._check_X_dict(X_dict)

        # calculating side and size from strategy pipeline
        signal_df = self.strategy_pipeline.get_side_and_size(X_dict)
        self._check_signal_has_valid_output(signal_df)

        return signal_df

    def create_strategy_bars(self, X_dict: Dict[str, pd.DataFrame]) -> pd.DataFrame:
        """
        Create the dataframe with the strategy bars information (stoploss,
        take profit and expiration)

        :param X_dict: A dictionary containing features for the primary and
                       meta-labeling model
        :type X_dict: Dict[str, pd.DataFrame]
        :return: dataframe with stop loss, taking profit and expiration
        :rtype: pd.DataFrame
        """
        self._check_X_dict(X_dict)

        # target dataframe is extracted so that the volatility can be
        # calculated
        strategy_bars_df = X_dict[self.strategy_pipeline.align_on_]
        volatility = get_daily_vol(strategy_bars_df[CLOSE], self.volatility_window)

        # adding stop loss and profit taking
        strategy_bars_df[STOP_LOSS] = volatility * self.stop_loss_factor
        strategy_bars_df[PROFIT_TAKING] = volatility * self.profit_taking_factor

        # adding number of expiration bars
        strategy_bars_df[NUMBER_EXPIRATION_BARS_COLUMN] = self.num_expiration_bars
        return strategy_bars_df[
            [STOP_LOSS, PROFIT_TAKING, NUMBER_EXPIRATION_BARS_COLUMN]
        ]

    def create_position(self, X_dict: Dict[str, pd.DataFrame]) -> pd.DataFrame:
        """
        Create a dataframe that can be used to evaluate the strategy.

        The dataframe contains close, stop_loss, profit_taking,
        number of expiration bars, posiion size and side.

        :param X_dict: A dictionary containing features for the primary and
                       meta-labeling model
        :type X_dict: Dict[str, pd.DataFrame]
        :return: The strategy positions and related informations
        """
        # TODO: compare the predictions of the primary with the metalabeling

        # calculating side and size from strategy pipeline
        signal_df = self.create_signal(X_dict)
        strategy_bars_df = self.create_strategy_bars(X_dict)

        position_df = pd.concat([signal_df, strategy_bars_df], axis=1)

        # check all the expected column are in position_df
        check_missing_columns(
            position_df,
            [STOP_LOSS, PROFIT_TAKING, NUMBER_EXPIRATION_BARS_COLUMN, SIZE, SIDE],
        )

        # Checking if position_df has unique indices
        if not position_df.index.is_unique:
            raise ValueError("The position dataframe has duplicated indices")

        return position_df



import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator
from sklearn.base import TransformerMixin


def _cast_nan_to_none(val):
    return val if not np.isnan(val) else None


def check_missing_columns(df, wanted_columns):
    assert set(df.columns).issuperset(
        set(wanted_columns)
    ), f"{set(wanted_columns) - set(df.columns)} is missing"


def convert_to_timestamp(datetime_array: np.ndarray) -> np.ndarray:
    """
    Convert a datetime array to timestamp in milliseconds.

    Nans are converted to negatives values

    :param datetime_array: An array of datetimes
    :return: array of timestamps
    :rtype: np.ndarray
    """

    return datetime_array.astype(int) // 10 ** 6


class IdentityTransformer(TransformerMixin, BaseEstimator):
    """The identity transformer returns exactly what the input is"""

    def fit(self, X, y=None, **fit_params):
        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        return X



import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator
from sklearn.base import TransformerMixin


class BuySellImbalance(BaseEstimator, TransformerMixin):
    allowed_volume_types = ["quote", "base"]

    def __init__(self, fast: int, slow: int, volume_type: str):
        self.fast = fast
        self.slow = slow
        if volume_type not in self.allowed_volume_types:
            raise ValueError("The allowed volume types are 'quote' and 'base'")
        self.volume_type = volume_type

    def fit(self, X, y=None, **fit_params):
        return self

    def transform(self, X: pd.DataFrame) -> np.ndarray:

        fast_buy_sell_diff = (
            X[f"{self.volume_type}_asset_buy_volume"].ewm(span=self.fast).mean()
            - X[f"{self.volume_type}_asset_sell_volume"].ewm(span=self.fast).mean()
        )
        slow_buy_sell_diff = (
            X[f"{self.volume_type}_asset_buy_volume"].ewm(span=self.slow).mean()
            - X[f"{self.volume_type}_asset_sell_volume"].ewm(span=self.slow).mean()
        )

        fast_sell_crossover_normalised = (fast_buy_sell_diff - slow_buy_sell_diff) / (
            X[f"{self.volume_type}_asset_volume"] + 1
        )

        return fast_sell_crossover_normalised.values.reshape(-1, 1)



from abc import abstractmethod

import numpy as np
import pandas as pd
from mizarlabs.static import EVENT_END_TIME
from mizarlabs.transformers.utils import check_missing_columns
from mizarlabs.transformers.utils import convert_to_timestamp
from numba import jit
from numba import prange
from scipy.stats import norm
from sklearn.base import BaseEstimator
from sklearn.base import TransformerMixin

BET_SIZE = "bet_size"
PROBABILITY = "prob"
PREDICTION = "pred"
SIDE = "side"


class BetSizingBase(BaseEstimator, TransformerMixin):
    """
    Base class for bet sizing transformers
    """

    def transform(self, X: pd.DataFrame) -> pd.Series:

        bet_sizing_signal = self._transform(X)

        assert (bet_sizing_signal >= -1).all(), "The bet size signal should be >= -1"
        assert (bet_sizing_signal <= 1).all(), "The bet size signal should be <= 1"

        return bet_sizing_signal

    @abstractmethod
    def _transform(self, X: pd.DataFrame) -> pd.Series:
        pass


class BetSizingFromProbabilities(BetSizingBase):
    """
    Calculate the bet size using the predicted probability.

    :param num_classes: Number of labeled classes
    :type num_classes: int
    :param average_active: Whether we need to apply the average
                           active to the bet sizing signal
    :type average_active: bool, optional
    :param meta_labeling: Whether the bet sizing is calculated from a
                          metalabeling signal
    :type meta_labeling: bool, optional
    :param discretise: Whether the output needs to be discretised
    :type discretise: bool, optional
    :param step_size: The step size of the discretisation
    :type step_size: int, optional
    :param probability_column_name: The column name of the probabilities
    :type probability_column_name: str, optional
    :param prediction_column_name: The column name of the predictions
    :type prediction_column_name: str, optional
    :param side_column_name: The column name of the side of the 'simpler'
                             metalabeling model
    :type side_column_name: str, optional
    :param event_end_time_column_name: The column name of the event end time
    :rtype event_end_time_column_name: str, optional
    """

    def __init__(
        self,
        num_classes: int,
        average_active: bool = False,
        meta_labeling: bool = False,
        discretise: bool = False,
        step_size: float = None,
        probability_column_name: str = PROBABILITY,
        prediction_column_name: str = PREDICTION,
        side_column_name: str = SIDE,
        event_end_time_column_name: str = EVENT_END_TIME,
        bet_size_column_name: str = BET_SIZE,
    ):

        self._side_column_name = side_column_name
        self._metalabeling = meta_labeling
        self._average_active = average_active
        self._step_size = step_size
        self._num_classes = num_classes
        self._probability_column_name = probability_column_name
        self._prediction_column_name = prediction_column_name
        self._event_end_time_column_name = event_end_time_column_name
        self._discretise = discretise
        self._bet_size_column_name = bet_size_column_name

        if self._discretise:
            assert self._discretise and self._step_size, (
                "When discretise is activated, step size should be "
                "set with value between 0 and 1"
            )
            assert (
                0 < self._step_size < 1
            ), "The step size should be greater than zero and less than 1"

    def _transform(self, X: pd.DataFrame) -> pd.Series:
        check_missing_columns(
            X, [self._probability_column_name, self._prediction_column_name]
        )

        # generate signals from multinomial classification (one-vs-rest, OvR)
        test_statistic_z = (
            X[self._probability_column_name] - 1.0 / self._num_classes
        ) / (
            X[self._probability_column_name] * (1.0 - X[self._probability_column_name])
        ) ** 0.5

        # signal=side*size
        bet_sizing_signal = X[self._prediction_column_name] * (
            2 * norm.cdf(test_statistic_z) - 1
        )

        if self._metalabeling:
            assert set(X[self._side_column_name].unique()).issubset(
                {1, -1, 0}
            ), "The side should be 1, -1 or 0"

            check_missing_columns(X, [self._side_column_name])
            bet_sizing_signal *= X.loc[bet_sizing_signal.index, self._side_column_name]

        if self._average_active:
            bet_sizing_signal_with_barrier = bet_sizing_signal.to_frame(BET_SIZE).join(
                X[[self._event_end_time_column_name]], how="left"
            )

            bet_sizing_signal = avg_active_signals(
                bet_sizing_signal_with_barrier,
                self._event_end_time_column_name,
                self._bet_size_column_name,
            )

        if self._discretise:
            bet_sizing_signal = discretise_signal(bet_sizing_signal, self._step_size)

        return bet_sizing_signal.abs()


def avg_active_signals(
    signals: pd.DataFrame,
    event_end_time_column_name: str = EVENT_END_TIME,
    bet_size_column_name: str = BET_SIZE,
) -> pd.Series:
    """
    Average the bet sizes of all concurrently not closed positions
    (e.i. no barrier has been hit yet)

    :param signals: Signal from which the active average is calculated
    :rtype signals: pd.DataFrame
    :param event_end_time_column_name: the name of the event end time
    :type event_end_time_column_name: str, optional
    :param bet_size_column_name: the name of the bet size column
    :type bet_size_column_name: str, optional
    :return: The active average signal
    :rtype: pd.DataFrame
    """
    # compute the average bet size among those active
    # time points were bet size change (either one starts or one ends)
    active_bet_size_time_indices = set(
        signals[event_end_time_column_name].dropna().values
    )
    active_bet_size_time_indices = active_bet_size_time_indices.union(
        signals.index.values
    )
    active_bet_size_time_indices = list(active_bet_size_time_indices)
    active_bet_size_time_indices.sort()
    active_bet_size_time_indices = np.array(active_bet_size_time_indices)

    avg_active_bet_size_list = _get_avg_active_signals(
        signals.loc[:, bet_size_column_name].values,
        convert_to_timestamp(active_bet_size_time_indices),
        convert_to_timestamp(signals.index.values),
        convert_to_timestamp(signals[event_end_time_column_name].values),
    )

    avg_active_bet_size = pd.Series(
        avg_active_bet_size_list, index=active_bet_size_time_indices, dtype=float
    )

    return avg_active_bet_size


@jit(parallel=True, nopython=True)
def _get_avg_active_signals(
    bet_size_signal: np.ndarray,
    active_bet_size_time_indices: np.ndarray,
    signal_timestamp_index: np.ndarray,
    expiration_barrier_timestamp: np.ndarray,
) -> np.ndarray:
    """
    Calculate the average active bet signal from the overlapping bets

    :param bet_size_signal: The bet size signal not averaged by active signals
    :type bet_size_signal: np.ndarray
    :param active_bet_size_time_indices: The timestamps when at least one
                                         signal is active
    :type active_bet_size_time_indices: np.ndarray
    :param signal_timestamp_index: The timestamps of the signal bet signal
    :type signal_timestamp_index: np.ndarray
    :param expiration_barrier_timestamp: The timestamps of the expiration
                                         barriers
    :type expiration_barrier_timestamp: np.ndarray
    :return: The average active bet size
    :rtype: np.ndarray
    """
    # init the average active bet sizes array with zeros
    avg_active_bet_size = np.zeros_like(active_bet_size_time_indices, dtype=np.float64)

    for i in prange(len(active_bet_size_time_indices)):
        active_bet_size_time = active_bet_size_time_indices[i]
        # mask that finds where the bet signals are overlapping
        mask = np.less_equal(
            signal_timestamp_index, active_bet_size_time
        ) * np.logical_or(
            np.less(active_bet_size_time, expiration_barrier_timestamp),
            np.less(expiration_barrier_timestamp, 0),
        )
        # select the active bet sizes signals and calculates the mean
        active_bets_timestamps = signal_timestamp_index[mask]
        if len(active_bets_timestamps) > 0:
            avg_active_bet_size[i] = np.mean(bet_size_signal[mask])

    return avg_active_bet_size


def discretise_signal(signal: pd.Series, step_size: float) -> pd.Series:
    """
    Discretise the bet size signal based on the step size given.

    :param signal: Signal to discretise
    :type signal: pd.Series
    :param step_size: the step size to use for the discretisation
    :type step_size: float
    :return: Discretised signal
    :rtype: pd.Series
    """
    assert 0 < step_size < 1, "The step size should be between 0 and 1"

    discretised_signal = ((signal / step_size).round() * step_size).round(3)
    # Capping the discretised signal to 1
    discretised_signal[discretised_signal > 1] = 1
    # Flooring the discretised signal to 0
    discretised_signal[discretised_signal < -1] = -1

    return discretised_signal
